/***************************************************************************
 $RCSfile: message.c,v $
                             -------------------
    cvs         : $Id: message.c,v 1.33 2005/01/30 14:47:35 aquamaniac Exp $
    begin       : Mon Mar 01 2004
    copyright   : (C) 2004 by Martin Preuss
    email       : martin@libchipcard.de

 ***************************************************************************
 *          Please see toplevel file COPYING for license details           *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif


/* #define AH_MSG_HEAVY_DEBUG */

#include "message_p.h"
#include "aqhbci_l.h"
#include "hbci_l.h"
#include <gwenhywfar/gwenhywfar.h>
#include <gwenhywfar/debug.h>
#include <gwenhywfar/misc.h>
#include <gwenhywfar/text.h>
#include <gwenhywfar/list.h>
#include <aqhbci/msgengine.h>
#include <aqhbci/mediumpintan.h>
#include <aqbanking/banking.h>

#include <gwenhywfar/version.h>
#include <gwenhywfar/directory.h>

#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>



GWEN_LIST_FUNCTIONS(AH_MSG, AH_Msg);



/* --------------------------------------------------------------- FUNCTION */
GWEN_KEYSPEC *AH_Msg_GetSigners(const AH_MSG *hmsg){
  assert(hmsg);
  return hmsg->signers;
}



/* --------------------------------------------------------------- FUNCTION */
int AH_Msg_AddSigner(AH_MSG *hmsg, const GWEN_KEYSPEC *ks){
  assert(hmsg);
  assert(ks);

  if (hmsg->nodes) {
    DBG_ERROR(AQHBCI_LOGDOMAIN, "Signers must be added before nodes !");
    return -1;
  }
  GWEN_KeySpec_Add(GWEN_KeySpec_dup(ks), &(hmsg->signers));
  hmsg->nSigners++;
  return 0;
}



/* --------------------------------------------------------------- FUNCTION */
unsigned int AH_Msg_GetSignerCount(AH_MSG *hmsg){
  assert(hmsg);
  return hmsg->nSigners;
}



/* --------------------------------------------------------------- FUNCTION */
const GWEN_KEYSPEC *AH_Msg_GetCrypter(const AH_MSG *hmsg){
  assert(hmsg);
  return hmsg->crypter;
}



/* --------------------------------------------------------------- FUNCTION */
void AH_Msg_SetCrypter(AH_MSG *hmsg, const GWEN_KEYSPEC *ks){
  assert(hmsg);
  GWEN_KeySpec_free(hmsg->crypter);
  hmsg->crypter=GWEN_KeySpec_dup(ks);
}



/* --------------------------------------------------------------- FUNCTION */
GWEN_BUFFER *AH_Msg_GetBuffer(AH_MSG *hmsg){
  assert(hmsg);
  return hmsg->buffer;
}



/* --------------------------------------------------------------- FUNCTION */
GWEN_BUFFER *AH_Msg_TakeBuffer(AH_MSG *hmsg){
  GWEN_BUFFER *bf;

  assert(hmsg);
  bf=hmsg->buffer;
  hmsg->buffer=0;
  return bf;
}



/* --------------------------------------------------------------- FUNCTION */
void AH_Msg_SetBuffer(AH_MSG *hmsg,
                           GWEN_BUFFER *bf){
  assert(hmsg);
  GWEN_Buffer_free(hmsg->buffer);
  hmsg->buffer=bf;
}



/* --------------------------------------------------------------- FUNCTION */
unsigned int AH_Msg_GetMsgNum(const AH_MSG *hmsg){
  assert(hmsg);
  return hmsg->msgNum;
}



/* --------------------------------------------------------------- FUNCTION */
unsigned int AH_Msg_GetMsgRef(const AH_MSG *hmsg){
  assert(hmsg);
  return hmsg->refMsgNum;
}



/* --------------------------------------------------------------- FUNCTION */
void AH_Msg_SetMsgRef(AH_MSG *hmsg, unsigned int i){
  assert(hmsg);
  hmsg->refMsgNum=i;
}



/* --------------------------------------------------------------- FUNCTION */
unsigned int AH_Msg_GetNodes(AH_MSG *hmsg){
  assert(hmsg);
  return hmsg->nodes;
}



/* --------------------------------------------------------------- FUNCTION */
AH_MSG *AH_Msg_new(AH_DIALOG *dlg){
  AH_MSG *hmsg;

  assert(dlg);
  GWEN_NEW_OBJECT(AH_MSG, hmsg);
  GWEN_LIST_INIT(AH_MSG, hmsg);
  hmsg->dialog=dlg;
  AH_Dialog_Attach(dlg);
  hmsg->buffer=GWEN_Buffer_new(0, AH_MSG_DEFAULTSIZE, 0, 1);
  GWEN_Buffer_ReserveBytes(hmsg->buffer, AH_MSG_DEFAULTRESERVE);
  GWEN_Buffer_SetStep(hmsg->buffer, AH_MSG_DEFAULTSTEP);

  return hmsg;
}



/* --------------------------------------------------------------- FUNCTION */
void AH_Msg_free(AH_MSG *hmsg){
  if (hmsg) {
    DBG_INFO(AQHBCI_LOGDOMAIN, "Destroying AH_MSG");
    GWEN_LIST_FINI(AH_MSG, hmsg);
    GWEN_KeySpec_Clear(&(hmsg->signers));
    GWEN_KeySpec_free(hmsg->crypter);
    GWEN_Buffer_free(hmsg->buffer);
    GWEN_Buffer_free(hmsg->origbuffer);
    AH_Dialog_free(hmsg->dialog);
    free(hmsg->resultText);
    free(hmsg->resultParam);
    free(hmsg->usedTan);
    free(hmsg->expectedSigner);
    free(hmsg->expectedCrypter);
    if (hmsg->usedPin) {
      memset(hmsg->usedPin, 0, strlen(hmsg->usedPin));
      free(hmsg->usedPin);
    }
    GWEN_DB_Group_free(hmsg->decodedMsg);

    free(hmsg);
  }
}



/* --------------------------------------------------------------- FUNCTION */
int AH_Msg_AddMsgTail(AH_MSG *hmsg){
  GWEN_XMLNODE *node;
  GWEN_DB_NODE *cfg;
  int rv;
  GWEN_MSGENGINE *e;

  assert(hmsg);
  e=AH_Dialog_GetMsgEngine(hmsg->dialog);
  assert(e);

  node=GWEN_MsgEngine_FindNodeByProperty(e,
                                         "SEG",
                                         "id",
                                         0,
                                         "MsgTail");
  if (!node) {
    DBG_INFO(AQHBCI_LOGDOMAIN, "Segment \"MsgTail\"not found");
    return -1;
  }

  cfg=GWEN_DB_Group_new("msgtail");
  GWEN_DB_SetIntValue(cfg, GWEN_DB_FLAGS_DEFAULT,
                      "head/seq",
                      hmsg->lastSegment+1);
  GWEN_Buffer_SetPos(hmsg->buffer, GWEN_Buffer_GetUsedBytes(hmsg->buffer));
  rv=GWEN_MsgEngine_CreateMessageFromNode(e,
                                          node,
                                          hmsg->buffer,
                                          cfg);
  GWEN_DB_Group_free(cfg);
  if (rv) {
    DBG_INFO(AQHBCI_LOGDOMAIN, "Could not create msgTail");
    return -1;
  }

  return 0;
}



/* --------------------------------------------------------------- FUNCTION */
int AH_Msg_AddMsgHead(AH_MSG *hmsg) {
  GWEN_XMLNODE *node;
  GWEN_DB_NODE *cfg;
  GWEN_BUFFER *hbuf;
  unsigned int msize;
  int rv;
  GWEN_MSGENGINE *e;

  assert(hmsg);
  e=AH_Dialog_GetMsgEngine(hmsg->dialog);
  assert(e);

  node=GWEN_MsgEngine_FindNodeByProperty(e,
                                         "SEG",
                                         "id",
                                         0,
                                         "MsgHead");
  if (!node) {
    DBG_INFO(AQHBCI_LOGDOMAIN, "Segment \"MsgHead\"not found");
    return -1;
  }

  cfg=GWEN_DB_Group_new("msghead");
  GWEN_DB_SetCharValue(cfg, GWEN_DB_FLAGS_DEFAULT,
                       "dialogid",
                       AH_Dialog_GetDialogId(hmsg->dialog));
  GWEN_DB_SetIntValue(cfg, GWEN_DB_FLAGS_DEFAULT,
                      "msgnum",
                      hmsg->msgNum);
  if (hmsg->refMsgNum) {
    DBG_INFO(AQHBCI_LOGDOMAIN, "Adding Reference Message Number");
    /* add message reference */
    GWEN_DB_SetIntValue(cfg, GWEN_DB_FLAGS_DEFAULT,
                        "msgref/msgnum", hmsg->refMsgNum);
    GWEN_DB_SetCharValue(cfg, GWEN_DB_FLAGS_DEFAULT,
                         "msgref/dialogid",
                         AH_Dialog_GetDialogId(hmsg->dialog));
  }
  DBG_INFO(AQHBCI_LOGDOMAIN, "Protocol version is %d",
           GWEN_MsgEngine_GetProtocolVersion(e));

  GWEN_DB_SetIntValue(cfg, GWEN_DB_FLAGS_DEFAULT,
                      "hversion",
                      GWEN_MsgEngine_GetProtocolVersion(e));
  GWEN_DB_SetIntValue(cfg, GWEN_DB_FLAGS_DEFAULT,
                      "size", 1);
  GWEN_DB_SetIntValue(cfg, GWEN_DB_FLAGS_DEFAULT,
                      "head/seq", 1);
  hbuf=GWEN_Buffer_new(0, 128, 0, 1);

  /* create first version of msgHead just to calculate the size */
  rv=GWEN_MsgEngine_CreateMessageFromNode(e,
                                          node,
                                          hbuf,
                                          cfg);
  if (rv) {
    DBG_INFO(AQHBCI_LOGDOMAIN, "Could not create msgHead");
    GWEN_Buffer_free(hbuf);
    GWEN_DB_Group_free(cfg);
    return -1;
  }

  msize=GWEN_Buffer_GetUsedBytes(hmsg->buffer)+
    GWEN_Buffer_GetUsedBytes(hbuf);
  DBG_INFO(AQHBCI_LOGDOMAIN, "Message size is: %d", msize);
  GWEN_DB_SetIntValue(cfg,
                      GWEN_DB_FLAGS_OVERWRITE_VARS,
                      "size", msize);
  GWEN_DB_SetIntValue(cfg, GWEN_DB_FLAGS_DEFAULT,
                      "head/seq", 1);
  GWEN_Buffer_Reset(hbuf);

  /* create final version of msgHead (we now know the size) */
  rv=GWEN_MsgEngine_CreateMessageFromNode(e,
                                          node,
                                          hbuf,
                                          cfg);
  GWEN_DB_Group_free(cfg);
  if (rv) {
    DBG_INFO(AQHBCI_LOGDOMAIN, "Could not create 2nd version of msgHead");
    GWEN_Buffer_free(hbuf);
    return -1;
  }

  /* insert msgHead */
  GWEN_Buffer_SetPos(hmsg->buffer, 0);
  if (GWEN_Buffer_InsertBuffer(hmsg->buffer, hbuf)) {
    DBG_INFO(AQHBCI_LOGDOMAIN, "Could not insert msgHead");
    GWEN_Buffer_free(hbuf);
    return -1;
  }

  GWEN_Buffer_free(hbuf);
  return 0;
}



/* --------------------------------------------------------------- FUNCTION */
unsigned int AH_Msg_GetCurrentSegmentNumber(AH_MSG *hmsg) {
  if (hmsg->firstSegment==0) {
    unsigned int rv;

    rv=2;
    if (hmsg->enableInsert)
      rv++;
    rv+=hmsg->nSigners;
    return rv;
  }
  return hmsg->lastSegment+1;
}



/* --------------------------------------------------------------- FUNCTION */
unsigned int AH_Msg_AddNode(AH_MSG *hmsg,
                            GWEN_XMLNODE *node,
                            GWEN_DB_NODE *data) {
  int rv;
  GWEN_MSGENGINE *e;
  unsigned int usedBefore;

  assert(hmsg);
  assert(node);
  assert(data);

  assert(hmsg);
  e=AH_Dialog_GetMsgEngine(hmsg->dialog);
  assert(e);

  if (hmsg->firstSegment==0) {
    /* first node */
    hmsg->firstSegment=2;
    if (hmsg->enableInsert)
      hmsg->firstSegment++;
    hmsg->firstSegment+=hmsg->nSigners;
    GWEN_MsgEngine_SetIntValue(e,
                               "SegmentNumber",
                               hmsg->firstSegment);
    hmsg->lastSegment=hmsg->firstSegment-1;
  }

  usedBefore=GWEN_Buffer_GetUsedBytes(hmsg->buffer);
  rv=GWEN_MsgEngine_CreateMessageFromNode(e,
                                          node,
                                          hmsg->buffer,
                                          data);
  if (rv) {
    DBG_INFO(AQHBCI_LOGDOMAIN, "here");
    GWEN_Buffer_Crop(hmsg->buffer, 0, usedBefore);
    GWEN_Buffer_SetPos(hmsg->buffer, usedBefore);
    return 0;
  }
  hmsg->lastSegment=GWEN_MsgEngine_GetIntValue(e,
                                               "SegmentNumber",
                                               1)-1;
  hmsg->nodes++;
  return hmsg->lastSegment;
}



/* --------------------------------------------------------------- FUNCTION */
unsigned int AH_Msg_InsertNode(AH_MSG *hmsg,
                               GWEN_XMLNODE *node,
                               GWEN_DB_NODE *data) {
  int rv;
  GWEN_MSGENGINE *e;
  GWEN_BUFFER *tmpbuf;
  unsigned int pos;

  assert(hmsg);
  assert(node);
  assert(data);

  if (hmsg->nodes==0)
    return AH_Msg_AddNode(hmsg, node, data);

  e=AH_Dialog_GetMsgEngine(hmsg->dialog);
  assert(e);

  if (!hmsg->enableInsert) {
    DBG_ERROR(AQHBCI_LOGDOMAIN, "Inserting a node without enableInsert flag");
    return 0;
  }
  if (hmsg->firstSegment<(2+hmsg->nSigners)) {
    DBG_ERROR(AQHBCI_LOGDOMAIN,
              "First segment is %d (%d signers), can't insert a new one",
              hmsg->firstSegment, hmsg->nSigners);
    return 0;
  }

  hmsg->firstSegment--;

  GWEN_MsgEngine_SetIntValue(e,
                             "SegmentNumber",
                             hmsg->firstSegment);

  tmpbuf=GWEN_Buffer_new(0, 512, 0, 1);
  rv=GWEN_MsgEngine_CreateMessageFromNode(e,
                                          node,
                                          tmpbuf,
                                          data);
  GWEN_MsgEngine_SetIntValue(e,
                             "SegmentNumber",
                             hmsg->lastSegment+1);
  if (rv) {
    DBG_INFO(AQHBCI_LOGDOMAIN, "here");
    GWEN_Buffer_free(tmpbuf);
    return 0;
  }

  pos=GWEN_Buffer_GetPos(hmsg->buffer);
  GWEN_Buffer_SetPos(hmsg->buffer, 0);
  GWEN_Buffer_InsertBuffer(hmsg->buffer, tmpbuf);
  GWEN_Buffer_SetPos(hmsg->buffer, pos+GWEN_Buffer_GetUsedBytes(tmpbuf));
  GWEN_Buffer_free(tmpbuf);

  hmsg->nodes++;
  return hmsg->firstSegment;
}



/* --------------------------------------------------------------- FUNCTION */
int AH_Msg_PrepareCryptoSeg(AH_MSG *hmsg,
                            GWEN_DB_NODE *cfg,
                            const GWEN_KEYSPEC *ks,
                            int crypt,
                            int createCtrlRef) {
  char sdate[9];
  char stime[7];
  char ctrlref[15];
  struct tm *lt;
  time_t tt;
  GWEN_MSGENGINE *e;
  AH_BANK *b;
  AH_USER *u;
  const char *userId;

  assert(hmsg);
  assert(cfg);
  assert(ks);

  b=AH_Dialog_GetBank(hmsg->dialog);
  assert(b);
  e=AH_Dialog_GetMsgEngine(hmsg->dialog);
  assert(e);
  u=AH_Customer_GetUser(AH_Dialog_GetDialogOwner(hmsg->dialog));
  assert(u);

  userId=GWEN_KeySpec_GetOwner(ks);
  if (!userId || *userId==0)
    userId=AH_User_GetUserId(u);
  assert(userId);
  assert(*userId);

  tt=time(0);
  lt=localtime(&tt);

  if (createCtrlRef) {
    /* create control reference */
    if (!strftime(ctrlref, sizeof(ctrlref),
                  "%Y%m%d%H%M%S", lt)) {
      DBG_INFO(AQHBCI_LOGDOMAIN, "CtrlRef string too long");
      return -1;
    }

    GWEN_DB_SetCharValue(cfg, GWEN_DB_FLAGS_DEFAULT,
                         "ctrlref", ctrlref);
  }
  /* create date */
  if (!strftime(sdate, sizeof(sdate),
                "%Y%m%d", lt)) {
    DBG_INFO(AQHBCI_LOGDOMAIN, "Date string too long");
    return -1;
  }
  /* create time */
  if (!strftime(stime, sizeof(stime),
                "%H%M%S", lt)) {
    DBG_INFO(AQHBCI_LOGDOMAIN, "Date string too long");
    return -1;
  }

  DBG_INFO(AQHBCI_LOGDOMAIN, "Date and Time: %s / %s",
           sdate, stime);

  if (AH_Dialog_GetFlags(hmsg->dialog) &
      AH_DIALOG_FLAGS_INITIATOR)
    GWEN_DB_SetIntValue(cfg, GWEN_DB_FLAGS_DEFAULT,
                        "SecDetails/dir", 1);
  else
    GWEN_DB_SetIntValue(cfg, GWEN_DB_FLAGS_DEFAULT,
                        "SecDetails/dir", 2);
  GWEN_DB_SetCharValue(cfg, GWEN_DB_FLAGS_DEFAULT,
                       "SecStamp/date", sdate);
  GWEN_DB_SetCharValue(cfg, GWEN_DB_FLAGS_DEFAULT,
                       "SecStamp/time", stime);
  GWEN_DB_SetCharValue(cfg, GWEN_DB_FLAGS_DEFAULT,
                       "key/bankcode",
                       AH_Bank_GetBankId(b));
  GWEN_DB_SetCharValue(cfg, GWEN_DB_FLAGS_DEFAULT,
                       "key/userid",
                       userId);
  GWEN_DB_SetCharValue(cfg, GWEN_DB_FLAGS_DEFAULT,
                       "key/keytype",
                       GWEN_KeySpec_GetKeyName(ks));
  GWEN_DB_SetIntValue(cfg, GWEN_DB_FLAGS_DEFAULT,
                      "key/keynum",
                      GWEN_KeySpec_GetNumber(ks));
  GWEN_DB_SetIntValue(cfg, GWEN_DB_FLAGS_DEFAULT,
                      "key/keyversion",
                      GWEN_KeySpec_GetVersion(ks));

  return 0;
}



/* --------------------------------------------------------------- FUNCTION */
int AH_Msg_SignMsg(AH_MSG *hmsg,
                   GWEN_BUFFER *rawBuf,
                   const GWEN_KEYSPEC *ks) {
  GWEN_XMLNODE *node;
  GWEN_DB_NODE *cfg;
  GWEN_BUFFER *sigbuf;
  GWEN_BUFFER *hbuf;
  unsigned int l;
  int rv;
  char ctrlref[15];
  const char *p;
  GWEN_MSGENGINE *e;
  AH_MEDIUM *medium;
  AH_BANK *b;
  AH_USER *u;
  const char *userId;

  assert(hmsg);
  b=AH_Dialog_GetBank(hmsg->dialog);
  u=AH_Customer_GetUser(AH_Dialog_GetDialogOwner(hmsg->dialog));
  assert(u);
  e=AH_Dialog_GetMsgEngine(hmsg->dialog);
  assert(e);

  userId=GWEN_KeySpec_GetOwner(ks);
  if (!userId || *userId==0)
    userId=AH_User_GetUserId(u);
  assert(userId);
  assert(*userId);

  /* get medium */
  DBG_INFO(AQHBCI_LOGDOMAIN, "Getting medium for signing");
  medium=AH_HBCI_GetMedium(AH_Bank_GetHbci(b),
                           AH_Dialog_GetDialogOwner(hmsg->dialog),
                           AH_CryptoActionSign);
  if (!medium) {
    DBG_INFO(AQHBCI_LOGDOMAIN,
             "Could not find medium for user \"%s\"",
             AH_User_GetUserId(u));
    return -1;
  }

  GWEN_MsgEngine_SetMode(e,
                         AH_Medium_TypeToName(AH_Medium_GetMediumType(medium)));

  node=GWEN_MsgEngine_FindNodeByProperty(e,
                                         "SEG",
                                         "id",
                                         0,
                                         "SigHead");
  if (!node) {
    DBG_INFO(AQHBCI_LOGDOMAIN, "Segment \"SigHead\" not found");
    return -1;
  }

  cfg=GWEN_DB_Group_new("sighead");

  /* get real keyspec */
  ks=AH_Medium_GetLocalSignKeySpec(medium);
  if (!ks) {
    DBG_ERROR(AQHBCI_LOGDOMAIN, "No keyspec for local sign key");
    GWEN_DB_Group_free(cfg);
    return -1;
  }
  DBG_INFO(AQHBCI_LOGDOMAIN, "Really signing with \"%s\"",
           userId);

  /* prepare config for segment */
  if (AH_Msg_PrepareCryptoSeg(hmsg, cfg, ks, 0, 1)) {
    DBG_INFO(AQHBCI_LOGDOMAIN, "here");
    GWEN_DB_Group_free(cfg);
    return -1;
  }
  DBG_ERROR(0, "ping");

  /* store security id */
  if (strcasecmp(GWEN_MsgEngine_GetMode(e), "RDH")==0) {
    const char *secid;
    const GWEN_KEYSPEC *eks;

    /* set expected signer */
    if (AH_Customer_GetBankSigns(AH_Dialog_GetDialogOwner(hmsg->dialog))) {
      eks=AH_Medium_GetRemoteSignKeySpec(medium);
      if (eks) {
        const char *remoteId;

        remoteId=GWEN_KeySpec_GetOwner(eks);
        if (!remoteId || *remoteId==0)
          remoteId=AH_User_GetUserId(u);
        assert(remoteId);
        assert(*remoteId);

        DBG_INFO(AQHBCI_LOGDOMAIN,
                 "Expecting \"%s\" to sign the response",
                 remoteId);
        AH_Msg_SetExpectedSigner(hmsg, remoteId);
      }
    }

    if (hmsg->noSysId)
      secid="0";
    else {
      secid=AH_Customer_GetSystemId(AH_Dialog_GetDialogOwner(hmsg->dialog));
      if (!secid)
        secid="0";
    }

    /* RDH mode */
    GWEN_DB_SetCharValue(cfg, GWEN_DB_FLAGS_DEFAULT,
                         "SecDetails/SecId",
                         secid);
  }
  else if (strcasecmp(GWEN_MsgEngine_GetMode(e), "DDV")==0) {
    const char *secid;
    const GWEN_KEYSPEC *eks;

    /* set expected signer */
    if (AH_Customer_GetBankSigns(AH_Dialog_GetDialogOwner(hmsg->dialog))) {
      eks=AH_Medium_GetRemoteSignKeySpec(medium);
      if (eks) {
        const char *remoteId;

        remoteId=GWEN_KeySpec_GetOwner(eks);
        if (!remoteId || *remoteId==0)
          remoteId=AH_User_GetUserId(u);
        assert(remoteId);
        assert(*remoteId);

        DBG_INFO(AQHBCI_LOGDOMAIN,
                 "Expecting \"%s\" to sign the response",
                 remoteId);
        AH_Msg_SetExpectedSigner(hmsg, remoteId);
      }
    }

    secid=AH_Medium_GetSecurityIdPtr(medium);
    if (!secid) {
      DBG_ERROR(AQHBCI_LOGDOMAIN,
                "No system Id on DDV medium, should not happen");
      GWEN_DB_Group_free(cfg);
      return -1;
    }

    /* DDV mode */
    GWEN_DB_SetBinValue(cfg, GWEN_DB_FLAGS_DEFAULT,
                        "SecDetails/SecId",
                        AH_Medium_GetSecurityIdPtr(medium),
                        AH_Medium_GetSecurityIdSize(medium));
  }
  else if (strcasecmp(GWEN_MsgEngine_GetMode(e), "PINTAN")==0) {
    const char *secid;

    if (hmsg->noSysId)
      secid="0";
    else {
      secid=AH_Customer_GetSystemId(AH_Dialog_GetDialogOwner(hmsg->dialog));
      if (!secid)
        secid="0";
    }

    /* PINTAN mode, uses a system id as RDH mode does */
    GWEN_DB_SetCharValue(cfg, GWEN_DB_FLAGS_DEFAULT,
                         "SecDetails/SecId",
                         secid);
  }
  else {
    DBG_ERROR(AQHBCI_LOGDOMAIN, "Unknown mode \"%s\"", GWEN_MsgEngine_GetMode(e));
    GWEN_DB_Group_free(cfg);
    return -1;
  }

  p=GWEN_DB_GetCharValue(cfg, "ctrlref", 0, "");
  if (strlen(p)>=sizeof(ctrlref)) {
    DBG_INFO(AQHBCI_LOGDOMAIN, "Control reference too long (14 bytes maximum)");
    GWEN_DB_Group_free(cfg);
    return -1;
  }
  strcpy(ctrlref, p);

  /* create SigHead */
  hbuf=GWEN_Buffer_new(0, 128+GWEN_Buffer_GetUsedBytes(rawBuf), 0, 1);
  GWEN_DB_SetIntValue(cfg, GWEN_DB_FLAGS_DEFAULT,
                      "head/seq", hmsg->firstSegment-1);
  GWEN_DB_SetIntValue(cfg, GWEN_DB_FLAGS_DEFAULT,
                      "signseq",
                      AH_Medium_GetNextSignSeq(medium));



  rv=GWEN_MsgEngine_CreateMessageFromNode(e,
                                          node,
                                          hbuf,
                                          cfg);
  GWEN_DB_Group_free(cfg);
  cfg=0;
  if (rv) {
    DBG_INFO(AQHBCI_LOGDOMAIN, "Could not create SigHead");
    GWEN_Buffer_free(hbuf);
    return -1;
  }

  /* remember size of sighead for now */
  l=GWEN_Buffer_GetUsedBytes(hbuf);

  /* add raw data to to-sign data buffer */
  if (GWEN_Buffer_AppendBuffer(hbuf, rawBuf)) {
    DBG_INFO(AQHBCI_LOGDOMAIN, "here");
    GWEN_Buffer_free(hbuf);
    return -1;
  }

  /* sign message, create sigtail */
  cfg=GWEN_DB_Group_new("sigtail");
  sigbuf=GWEN_Buffer_new(0, 512, 0, 1);
  if (strcasecmp(GWEN_MsgEngine_GetMode(e), "PINTAN")==0) {
    char pin[8];

    GWEN_Buffer_AppendString(sigbuf, "NOSIGNATURENEEDED");
    memset(pin, 0, sizeof(pin));
    if (AH_MediumPINTAN_GetPin(medium, pin, 4, sizeof(pin), 0)) {
      DBG_ERROR(AQHBCI_LOGDOMAIN, "Error getting pin from medium");
      GWEN_Buffer_free(sigbuf);
      GWEN_DB_Group_free(cfg);
      GWEN_Buffer_free(hbuf);
      return -1;
    }
    GWEN_DB_SetCharValue(cfg, GWEN_DB_FLAGS_DEFAULT,
                         "pin", pin);
    AH_Msg_SetPin(hmsg, pin);

    if (hmsg->needTan) {
      DBG_NOTICE(AQHBCI_LOGDOMAIN,
                 "This queue needs a TAN");
      if (hmsg->usedTan) {
        DBG_NOTICE(AQHBCI_LOGDOMAIN,
                   "Using existing TAN");
        GWEN_DB_SetCharValue(cfg, GWEN_DB_FLAGS_DEFAULT,
                             "tan", hmsg->usedTan);
      }
      else {
        char tan[16];

        memset(tan, 0, sizeof(tan));
        DBG_NOTICE(AQHBCI_LOGDOMAIN,
                   "Asking for TAN");
        if (AH_MediumPINTAN_GetTan(medium, tan, 4, sizeof(tan))) {
          DBG_ERROR(AQHBCI_LOGDOMAIN, "Error getting TAN from medium");
          GWEN_Buffer_free(sigbuf);
          GWEN_DB_Group_free(cfg);
          GWEN_Buffer_free(hbuf);
          return -1;
        }
        GWEN_DB_SetCharValue(cfg, GWEN_DB_FLAGS_DEFAULT,
                             "tan", tan);
        AH_Msg_SetTan(hmsg, tan);
      }
    }
    else {
      DBG_NOTICE(AQHBCI_LOGDOMAIN,
                 "This queue doesn't need a TAN");
    }
  }
  else {
    if (AH_Medium_Sign(medium, hbuf, sigbuf)!=AH_MediumResultOk) {
      DBG_INFO(AQHBCI_LOGDOMAIN,
               "Error signing with medium of user \"%s\"",
               userId);
      GWEN_Buffer_free(sigbuf);
      GWEN_DB_Group_free(cfg);
      GWEN_Buffer_free(hbuf);
      return -1;
    }
  }
  DBG_INFO(AQHBCI_LOGDOMAIN, "Signing done");

  /* insert new SigHead at beginning of message buffer */
  DBG_INFO(AQHBCI_LOGDOMAIN, "Inserting signature head");
  GWEN_Buffer_Rewind(hmsg->buffer);
  GWEN_Buffer_InsertBytes(hmsg->buffer, GWEN_Buffer_GetStart(hbuf), l);

  /* create sigtail */
  DBG_INFO(AQHBCI_LOGDOMAIN, "Completing signature tail");
  GWEN_Buffer_Reset(hbuf);
  GWEN_DB_SetIntValue(cfg, GWEN_DB_FLAGS_DEFAULT,
                      "head/seq", hmsg->lastSegment+1);
  /* store to DB */
  GWEN_DB_SetBinValue(cfg, GWEN_DB_FLAGS_DEFAULT,
                      "signature",
                      GWEN_Buffer_GetStart(sigbuf),
                      GWEN_Buffer_GetUsedBytes(sigbuf));

  GWEN_DB_SetCharValue(cfg, GWEN_DB_FLAGS_DEFAULT,
                       "ctrlref", ctrlref);
  GWEN_Buffer_free(sigbuf);

  /* get node */
  node=GWEN_MsgEngine_FindNodeByProperty(e,
                                         "SEG",
                                         "id",
                                         0,
                                         "SigTail");
  if (!node) {
    DBG_INFO(AQHBCI_LOGDOMAIN, "Segment \"SigTail\"not found");
    GWEN_Buffer_free(hbuf);
    GWEN_DB_Group_free(cfg);
    return -1;
  }
  rv=GWEN_MsgEngine_CreateMessageFromNode(e,
                                          node,
                                          hbuf,
                                          cfg);
  if (rv) {
    DBG_INFO(AQHBCI_LOGDOMAIN, "Could not create SigTail");
    GWEN_Buffer_free(hbuf);
    GWEN_DB_Group_free(cfg);
    return -1;
  }

  /* append sigtail */
  DBG_INFO(AQHBCI_LOGDOMAIN, "Appending signature tail");
  if (GWEN_Buffer_AppendBuffer(hmsg->buffer, hbuf)) {
    DBG_INFO(AQHBCI_LOGDOMAIN, "here");
    GWEN_Buffer_free(hbuf);
    GWEN_DB_Group_free(cfg);
    return -1;
  }
  DBG_INFO(AQHBCI_LOGDOMAIN, "Appending signature tail: done");

  GWEN_Buffer_free(hbuf);
  GWEN_DB_Group_free(cfg);

  /* adjust segment numbers (for next signature and message tail */
  hmsg->firstSegment--;
  hmsg->lastSegment++;

  return 0;
}



/* --------------------------------------------------------------- FUNCTION */
int AH_Msg_EncryptMsg(AH_MSG *hmsg) {
  GWEN_XMLNODE *node;
  GWEN_DB_NODE *cfg;
  GWEN_BUFFER *cryptbuf;
  GWEN_BUFFER *hbuf;
  GWEN_BUFFER *msgkeybuf;
  int rv;
  GWEN_MSGENGINE *e;
  AH_MEDIUM *medium;
  const GWEN_KEYSPEC *ks;
  AH_BANK *b;
  AH_USER *u;
  const char *localId;

  assert(hmsg);
  b=AH_Dialog_GetBank(hmsg->dialog);
  assert(b);
  u=AH_Customer_GetUser(AH_Dialog_GetDialogOwner(hmsg->dialog));
  assert(u);

  DBG_INFO(AQHBCI_LOGDOMAIN, "Encrypting message");
  e=AH_Dialog_GetMsgEngine(hmsg->dialog);
  assert(e);

  ks=hmsg->crypter;
  if (!ks) {
    DBG_ERROR(AQHBCI_LOGDOMAIN, "No crypter");
    return -1;
  }

  localId=GWEN_KeySpec_GetOwner(ks);
  if (!localId || *localId==0)
    localId=AH_User_GetUserId(u);
  assert(localId);
  assert(*localId);

  /* get medium */
  DBG_INFO(AQHBCI_LOGDOMAIN, "Getting medium for encrypting");
  medium=AH_HBCI_GetMedium(AH_Bank_GetHbci(b),
                           AH_Dialog_GetDialogOwner(hmsg->dialog),
                           AH_CryptoActionEncrypt);
  if (!medium) {
    DBG_INFO(AQHBCI_LOGDOMAIN,
             "Could not find medium for user \"%s\"",
             localId);
    /* GWEN_DB_Group_free(cfg); */
    return -1;
  }
  GWEN_MsgEngine_SetMode(e,
                         AH_Medium_TypeToName(AH_Medium_GetMediumType(medium)));

  node=GWEN_MsgEngine_FindNodeByProperty(e,
                                         "SEG",
                                         "id",
                                         0,
                                         "CryptHead");
  if (!node) {
    DBG_INFO(AQHBCI_LOGDOMAIN, "Segment \"CryptHead\"not found");
    return -1;
  }

  /* get real keyspec */
  ks=AH_Medium_GetRemoteCryptKeySpec(medium);
  if (!ks) {
    DBG_ERROR(AQHBCI_LOGDOMAIN, "No keyspec for remote crypt key");
    /* GWEN_DB_Group_free(cfg); */
    return -1;
  }

  localId=GWEN_KeySpec_GetOwner(ks);
  if (!localId || *localId==0)
    localId=AH_User_GetUserId(u);
  assert(localId);
  assert(*localId);

  /* encrypt message */
  DBG_INFO(AQHBCI_LOGDOMAIN, "Encrypting message");
  msgkeybuf=GWEN_Buffer_new(0, 128, 0, 1);
  cryptbuf=GWEN_Buffer_new(0, GWEN_Buffer_GetUsedBytes(hmsg->buffer)+256,0,1);
  GWEN_Buffer_Rewind(hmsg->buffer);

  if (strcasecmp(GWEN_MsgEngine_GetMode(e), "PINTAN")==0) {
    GWEN_Buffer_AppendBuffer(cryptbuf, hmsg->buffer);
    GWEN_Buffer_AppendString(msgkeybuf, "NOKEYINSSLMODE");
  }
  else {
    /* set expected crypter */
    DBG_INFO(AQHBCI_LOGDOMAIN,
             "Expecting \"%s\" to encrypt the response",
             localId);
    AH_Msg_SetExpectedCrypter(hmsg, localId);

    if (AH_Medium_Encrypt(medium, hmsg->buffer, cryptbuf, msgkeybuf)!=
        AH_MediumResultOk) {
      DBG_INFO(AQHBCI_LOGDOMAIN,
               "Error encrypting with medium for user \"%s\"",
               localId);
      GWEN_Buffer_free(cryptbuf);
      GWEN_Buffer_free(msgkeybuf);
      /* GWEN_DB_Group_free(cfg); */
      return -1;
    }
  }
  DBG_INFO(AQHBCI_LOGDOMAIN, "Encrypting message: done");

  /* create CryptHead */
  hbuf=GWEN_Buffer_new(0, 256+GWEN_Buffer_GetUsedBytes(hmsg->buffer), 0, 1);
  cfg=GWEN_DB_Group_new("crypthead");
  GWEN_DB_SetIntValue(cfg, GWEN_DB_FLAGS_DEFAULT,
                      "head/seq", 998);

  if (AH_Msg_PrepareCryptoSeg(hmsg, cfg, ks, 1, 0)) {
    DBG_INFO(AQHBCI_LOGDOMAIN, "here");
    GWEN_Buffer_free(hbuf);
    GWEN_DB_Group_free(cfg);
    return -1;
  }

  /* store security id */
  if (strcasecmp(GWEN_MsgEngine_GetMode(e), "RDH")==0) {
    const char *secid;

    if (hmsg->noSysId)
      secid="0";
    else {
      secid=AH_Customer_GetSystemId(AH_Dialog_GetDialogOwner(hmsg->dialog));
      if (!secid)
        secid="0";
    }

    /* RDH mode */
    GWEN_DB_SetCharValue(cfg, GWEN_DB_FLAGS_DEFAULT,
                         "SecDetails/SecId", secid);
  }
  else if (strcasecmp(GWEN_MsgEngine_GetMode(e), "DDV")==0) {
    const char *secid;

    secid=AH_Medium_GetSecurityIdPtr(medium);
    if (!secid) {
      DBG_ERROR(AQHBCI_LOGDOMAIN, "No system Id on DDV medium, should not happen");
      GWEN_DB_Group_free(cfg);
      return -1;
    }

    /* DDV mode */
    GWEN_DB_SetBinValue(cfg, GWEN_DB_FLAGS_DEFAULT,
                        "SecDetails/SecId",
                        AH_Medium_GetSecurityIdPtr(medium),
                        AH_Medium_GetSecurityIdSize(medium));
  }
  else if (strcasecmp(GWEN_MsgEngine_GetMode(e), "PINTAN")==0) {
    const char *secid;

    if (hmsg->noSysId)
      secid="0";
    else {
      secid=AH_Customer_GetSystemId(AH_Dialog_GetDialogOwner(hmsg->dialog));
      if (!secid)
        secid="0";
    }

    /* RDH mode */
    GWEN_DB_SetCharValue(cfg, GWEN_DB_FLAGS_DEFAULT,
                         "SecDetails/SecId",
                         secid);
  }
  else {
    DBG_ERROR(AQHBCI_LOGDOMAIN, "Unknown security mode \"%s\"", GWEN_MsgEngine_GetMode(e));
    GWEN_DB_Group_free(cfg);
    return -1;
  }

  /* store encrypted message key */
  GWEN_DB_SetBinValue(cfg, GWEN_DB_FLAGS_DEFAULT,
                      "CryptAlgo/MsgKey",
                      GWEN_Buffer_GetStart(msgkeybuf),
                      GWEN_Buffer_GetUsedBytes(msgkeybuf));
  GWEN_Buffer_free(msgkeybuf);

  rv=GWEN_MsgEngine_CreateMessageFromNode(e,
                                          node,
                                          hbuf,
                                          cfg);
  if (rv) {
    DBG_INFO(AQHBCI_LOGDOMAIN, "Could not create CryptHead");
    GWEN_Buffer_free(hbuf);
    GWEN_DB_Group_free(cfg);
    return -1;
  }

  GWEN_DB_Group_free(cfg);


  /* create cryptdata */
  cfg=GWEN_DB_Group_new("cryptdata");
  GWEN_DB_SetIntValue(cfg, GWEN_DB_FLAGS_DEFAULT,
                      "head/seq", 999);
  GWEN_DB_SetBinValue(cfg, GWEN_DB_FLAGS_DEFAULT,
                      "cryptdata",
                      GWEN_Buffer_GetStart(cryptbuf),
                      GWEN_Buffer_GetUsedBytes(cryptbuf));
  GWEN_Buffer_free(cryptbuf);

  /* get node */
  node=GWEN_MsgEngine_FindNodeByProperty(e,
                                         "SEG",
                                         "id",
                                         0,
                                         "CryptData");
  if (!node) {
    DBG_INFO(AQHBCI_LOGDOMAIN, "Segment \"CryptData\"not found");
    GWEN_Buffer_free(hbuf);
    GWEN_DB_Group_free(cfg);
    return -1;
  }
  rv=GWEN_MsgEngine_CreateMessageFromNode(e,
                                          node,
                                          hbuf,
                                          cfg);
  if (rv) {
    DBG_INFO(AQHBCI_LOGDOMAIN, "Could not create CryptData");
    GWEN_Buffer_free(hbuf);
    GWEN_DB_Group_free(cfg);
    return -1;
  }

  /* replace existing buffer by encrypted one */
  GWEN_Buffer_free(hmsg->buffer);
  hmsg->buffer=hbuf;
  GWEN_DB_Group_free(cfg);

  return 0;
}



/* --------------------------------------------------------------- FUNCTION */
int AH_Msg_EncodeMsg(AH_MSG *hmsg) {
  GWEN_MSGENGINE *e;

  assert(hmsg);

  e=AH_Dialog_GetMsgEngine(hmsg->dialog);
  assert(e);

  GWEN_MsgEngine_SetProtocolVersion(e, hmsg->hbciVersion);

  if (hmsg->firstSegment==0) {
    /* first node */
    hmsg->firstSegment=2;
    hmsg->firstSegment+=hmsg->nSigners;
    GWEN_MsgEngine_SetIntValue(e,
                               "SegmentNumber",
                               hmsg->firstSegment);
    hmsg->lastSegment=hmsg->firstSegment-1;
  }

  hmsg->msgNum=AH_Dialog_GetNextMsgNum(hmsg->dialog);
  GWEN_MsgEngine_SetIntValue(AH_Dialog_GetMsgEngine(hmsg->dialog),
                             "MessageNumber",
                             hmsg->msgNum);

  GWEN_MsgEngine_SetValue(AH_Dialog_GetMsgEngine(hmsg->dialog),
                          "DialogId",
                          AH_Dialog_GetDialogId(hmsg->dialog));

  /* sign message */
  DBG_INFO(AQHBCI_LOGDOMAIN, "Letting all signers sign");
  if (hmsg->nSigners) {
    GWEN_BUFFER *rawBuf;
    const GWEN_KEYSPEC *ks;

    rawBuf=GWEN_Buffer_dup(hmsg->buffer);
    ks=hmsg->signers;
    while (ks) {
      /*DBG_INFO(AQHBCI_LOGDOMAIN, "Signing with this key:");
       GWEN_KeySpec_Dump(ks, stderr, 1);*/
      if (AH_Msg_SignMsg(hmsg, rawBuf, ks)) {
        GWEN_Buffer_free(rawBuf);
        DBG_INFO(AQHBCI_LOGDOMAIN, "here");
        return -1;
      }
      ks=GWEN_KeySpec_ConstNext(ks);
    } /* while */
    GWEN_Buffer_free(rawBuf);
  } /* if signing is needed */
  DBG_INFO(AQHBCI_LOGDOMAIN, "Letting all signers sign: done");

  /* log unencrypted message */
  AH_Msg_LogMessage(hmsg, hmsg->buffer, 0, 0);

  /* encrypt message */
  if (hmsg->crypter) {
    DBG_INFO(AQHBCI_LOGDOMAIN, "Encrypting message");
    if (AH_Msg_EncryptMsg(hmsg)) {
      DBG_INFO(AQHBCI_LOGDOMAIN, "here");
      return -1;
    }
    DBG_INFO(AQHBCI_LOGDOMAIN, "Encrypting message: done");
  }

  /* add msg tail */
  DBG_INFO(AQHBCI_LOGDOMAIN, "Adding message tail");
  if (AH_Msg_AddMsgTail(hmsg)) {
    DBG_INFO(AQHBCI_LOGDOMAIN, "here");
    return -1;
  }
  DBG_INFO(AQHBCI_LOGDOMAIN, "Adding message tail: done");

  /* add msg head */
  DBG_INFO(AQHBCI_LOGDOMAIN, "Adding message head");
  if (AH_Msg_AddMsgHead(hmsg)) {
    DBG_INFO(AQHBCI_LOGDOMAIN, "here");
    return -1;
  }
  DBG_INFO(AQHBCI_LOGDOMAIN, "Adding message head: done");

  /* log final message */
  AH_Msg_LogMessage(hmsg, hmsg->buffer, 0, 1);

  DBG_INFO(AQHBCI_LOGDOMAIN, "Message finished");
  return 0;
}














/* --------------------------------------------------------------- FUNCTION */
/* return -1 on error (with group "seg/error" set) or -2 if the message is
 * faulty */
int AH_Msg_ReadSegment(AH_MSG *msg,
                       GWEN_MSGENGINE *e,
                       const char *gtype,
                       GWEN_BUFFER *mbuf,
                       GWEN_DB_NODE *gr,
                       unsigned int flags) {
  GWEN_XMLNODE *node;
  unsigned int posBak;
  const char *p;
  GWEN_DB_NODE *tmpdb;
  int segVer;

  /* find head segment description */
  tmpdb=GWEN_DB_Group_new("head");
  node=GWEN_MsgEngine_FindGroupByProperty(e,
                                          "id",
                                          0,
                                          "SegHead");
  if (node==0) {
    DBG_ERROR(AQHBCI_LOGDOMAIN, "Segment description not found (internal error)");
    GWEN_DB_Group_free(tmpdb);
    return -2;
  }

  /* parse head segment */
  posBak=GWEN_Buffer_GetPos(mbuf);
  if (GWEN_MsgEngine_ParseMessage(e,
                                  node,
                                  mbuf,
                                  tmpdb,
                                  flags)) {
    DBG_ERROR(AQHBCI_LOGDOMAIN, "Error parsing segment head");
    GWEN_DB_Group_free(tmpdb);
    return -2;
  }

  GWEN_Buffer_SetPos(mbuf, posBak);

  /* get segment code */
  segVer=GWEN_DB_GetIntValue(tmpdb,
                             "version",
                             0,
                             0);
  p=GWEN_DB_GetCharValue(tmpdb,
                         "code",
                         0,
                         0);
  if (!p) {
    DBG_ERROR(AQHBCI_LOGDOMAIN,
              "No segment code for %s ? This seems to be a bad msg...",
              gtype);
    DBG_ERROR(AQHBCI_LOGDOMAIN, "Full message (pos=%04x)", posBak);
    GWEN_Text_DumpString(GWEN_Buffer_GetStart(mbuf),
                         GWEN_Buffer_GetUsedBytes(mbuf),
                         stderr, 1);
    GWEN_DB_Dump(tmpdb, stderr, 1);
    GWEN_DB_Group_free(tmpdb);
    return -1;
  }

  /* try to find corresponding XML node */
  node=GWEN_MsgEngine_FindNodeByProperty(e,
                                         gtype,
                                         "code",
                                         segVer,
                                         p);
  if (node==0) {
    GWEN_DB_NODE *storegrp;
    unsigned int startPos;

    GWEN_Buffer_SetPos(mbuf, posBak);
    startPos=posBak;

    storegrp=GWEN_DB_GetGroup(gr,
                              GWEN_PATH_FLAGS_CREATE_GROUP,
                              p);
    assert(storegrp);
    GWEN_DB_AddGroup(storegrp, GWEN_DB_Group_dup(tmpdb));

    /* store the start position of this segment within the DB */
    GWEN_DB_SetIntValue(storegrp,
                        GWEN_DB_FLAGS_OVERWRITE_VARS,
                        "segment/pos",
                        startPos);
    GWEN_DB_SetIntValue(storegrp,
                        GWEN_DB_FLAGS_OVERWRITE_VARS,
                        "segment/error/code",
                        9130);
    GWEN_DB_SetCharValue(storegrp,
                         GWEN_DB_FLAGS_OVERWRITE_VARS,
                         "segment/error/text",
                         "Unbekanntes segment (Parser)");
    GWEN_DB_SetIntValue(storegrp,
                        GWEN_DB_FLAGS_OVERWRITE_VARS,
                        "segment/error/pos",
                        startPos);

    /* node not found, skip it */
    DBG_WARN(AQHBCI_LOGDOMAIN,
             "Unknown segment \"%s\" (Segnum=%d, version=%d, ref=%d)",
             p,
             GWEN_DB_GetIntValue(tmpdb, "seq", 0, -1),
             GWEN_DB_GetIntValue(tmpdb, "version", 0, -1),
             GWEN_DB_GetIntValue(tmpdb, "ref", 0, -1));
    if (GWEN_MsgEngine_SkipSegment(e, mbuf, '?', '\'')) {
      DBG_ERROR(AQHBCI_LOGDOMAIN, "Error skipping segment \"%s\"", p);
      GWEN_DB_Group_free(tmpdb);
      return -1;
    }
    /* store segment size within DB */
    GWEN_DB_SetIntValue(storegrp,
                        GWEN_DB_FLAGS_OVERWRITE_VARS,
                        "segment/length",
                        GWEN_Buffer_GetPos(mbuf)-startPos);
    /* handle trust info */
    if (flags & GWEN_MSGENGINE_READ_FLAGS_TRUSTINFO) {
      unsigned int usize;

      usize=GWEN_Buffer_GetPos(mbuf)-(startPos+1)-1;
      if (GWEN_MsgEngine_AddTrustInfo(e,
                                      GWEN_Buffer_GetStart(mbuf)+startPos,
                                      usize,
                                      p,
                                      GWEN_MsgEngineTrustLevelHigh,
                                      startPos)) {
        DBG_INFO(AQHBCI_LOGDOMAIN, "called from here");
        GWEN_DB_Group_free(tmpdb);
        return -1;
      }
    } /* if trustInfo handling wanted */
  }
  else {
    /* ok, node available, get the corresponding description and parse
     * the segment */
    const char *id;
    GWEN_DB_NODE *storegrp;
    unsigned int startPos;

    /* restore start position, since the segment head is part of a full
     * description, so we need to restart reading from the very begin */
    GWEN_Buffer_SetPos(mbuf, posBak);

    /* create group in DB for this segment */
    id=GWEN_XMLNode_GetProperty(node, "id", p);
    storegrp=GWEN_DB_GetGroup(gr,
                              GWEN_PATH_FLAGS_CREATE_GROUP,
                              id);
    assert(storegrp);

    /* store the start position of this segment within the DB */
    startPos=GWEN_Buffer_GetPos(mbuf);
    GWEN_DB_SetIntValue(storegrp,
                        GWEN_DB_FLAGS_OVERWRITE_VARS,
                        "segment/pos",
                        startPos);

    /* parse the segment */
    if (GWEN_MsgEngine_ParseMessage(e,
                                    node,
                                    mbuf,
                                    storegrp,
                                    flags)) {
      GWEN_DB_SetIntValue(storegrp,
                          GWEN_DB_FLAGS_OVERWRITE_VARS,
                          "segment/error/code",
                          9130);
      GWEN_DB_SetCharValue(storegrp,
                           GWEN_DB_FLAGS_OVERWRITE_VARS,
                           "segment/error/text",
                           "Syntaxfehler");
      GWEN_DB_SetIntValue(storegrp,
                          GWEN_DB_FLAGS_OVERWRITE_VARS,
                          "segment/error/pos",
                          GWEN_Buffer_GetPos(mbuf)-startPos);

      DBG_ERROR(AQHBCI_LOGDOMAIN, "Error parsing segment \"%s\"",p);
      GWEN_Text_DumpString(GWEN_Buffer_GetStart(mbuf)+startPos,
                           GWEN_Buffer_GetUsedBytes(mbuf)-startPos,
                           stderr, 1);
      GWEN_DB_Group_free(tmpdb);
      return -1;
    }

    /* store segment size within DB */
    GWEN_DB_SetIntValue(storegrp,
                        GWEN_DB_FLAGS_OVERWRITE_VARS,
                        "segment/length",
                        GWEN_Buffer_GetPos(mbuf)-startPos);
    if (strcasecmp(id, "MsgHead")==0) {
      int protocolVersion;

      protocolVersion=GWEN_DB_GetIntValue(storegrp,
                                          "hversion",
                                          0, 0);
      if (!protocolVersion) {
        DBG_WARN(AQHBCI_LOGDOMAIN, "Unknown protocol version, using default of 210");
        protocolVersion=210;
      }
      /* set HBCI protocol version to be used (taken from MsgHead) */
      DBG_INFO(AQHBCI_LOGDOMAIN, "Setting protocol version to %d", protocolVersion);
      GWEN_MsgEngine_SetProtocolVersion(e, protocolVersion);
      msg->hbciVersion=protocolVersion;
    }
  } /* if node found */
  GWEN_DB_Group_free(tmpdb);

  return 0;
}



/* --------------------------------------------------------------- FUNCTION */
int AH_Msg_ReadMessage(AH_MSG *msg,
                       GWEN_MSGENGINE *e,
                       const char *gtype,
                       GWEN_BUFFER *mbuf,
                       GWEN_DB_NODE *gr,
                       unsigned int flags) {
  unsigned int segments;
  unsigned int errors;
  int rv;

  segments=0;
  errors=0;

  while(GWEN_Buffer_GetBytesLeft(mbuf)) {
    rv=AH_Msg_ReadSegment(msg, e, gtype, mbuf, gr, flags);
    if (rv==-2) {
      DBG_INFO(AQHBCI_LOGDOMAIN, "here");
      return -1;
    }
    else if (rv==-1) {
      DBG_INFO(AQHBCI_LOGDOMAIN, "here");

      DBG_ERROR(AQHBCI_LOGDOMAIN, "Error here:");
      GWEN_Buffer_Dump(mbuf, stderr, 2);
#ifdef AH_MSG_HEAVY_DEBUG
      return -1;
#endif
      errors++;
    }
    segments++;
  } /* while */

  DBG_INFO(AQHBCI_LOGDOMAIN, "Parsed %d segments (%d had errors)",
           segments, errors);
  return 0;
}



/* --------------------------------------------------------------- FUNCTION */
int AH_Msg_PrepareCryptoSegDec(AH_MSG *hmsg,
                               GWEN_DB_NODE *n,
                               int crypt,
                               GWEN_KEYSPEC **keySpec,
                               int *signseq,
                               const char **pSecurityId,
                               int *lSecurityId,
                               GWEN_BUFFER *msgkeybuf) {
  GWEN_KEYSPEC *ks;
  const void *p;
  const char *s;
  unsigned int size;
  GWEN_MSGENGINE *e;
  int rdhMode;
  int pinTanMode;
  AH_BANK *b;

  assert(hmsg);
  assert(n);

  pinTanMode=0;
  rdhMode=0;
  b=AH_Dialog_GetBank(hmsg->dialog);
  e=AH_Dialog_GetMsgEngine(hmsg->dialog);
  assert(e);
  ks=GWEN_KeySpec_new();

  /* prepare keyspec */
  GWEN_KeySpec_SetOwner(ks, GWEN_DB_GetCharValue(n, "key/userid", 0, ""));
  GWEN_KeySpec_SetNumber(ks, GWEN_DB_GetIntValue(n, "key/keynum", 0, 0));
  GWEN_KeySpec_SetKeyName(ks, GWEN_DB_GetCharValue(n, "key/keytype", 0, ""));
  GWEN_KeySpec_SetVersion(ks,
                          GWEN_DB_GetIntValue(n, "key/keyversion", 0, 0));
  s=GWEN_DB_GetCharValue(n, "key/keytype", 0, 0);
  if (!s) {
    DBG_ERROR(AQHBCI_LOGDOMAIN, "no keytype");
    GWEN_KeySpec_free(ks);
    return -1;
  }
  GWEN_KeySpec_SetKeyName(ks, s);

  *keySpec=ks;

  if (GWEN_Text_Compare(GWEN_DB_GetCharValue(n, "key/bankcode", 0, ""),
                        AH_Bank_GetBankId(b),
                        1)) {
    DBG_WARN(AQHBCI_LOGDOMAIN,
             "BankId in received message does not match (%s!=%s)",
             GWEN_DB_GetCharValue(n, "key/bankcode", 0, ""),
             AH_Bank_GetBankId(b));
    AB_Banking_ProgressLog(AH_Dialog_GetBankingApi(hmsg->dialog),
                           0,
                           AB_Banking_LogLevelWarn,
                           I18N("BankId in received message does not match, "
                                "ignoring"));
  }

  /* get security mode */
  if (crypt) {
    int m;

    m=GWEN_DB_GetIntValue(n, "function", 0, -1);
    if (m==-1) {
      DBG_ERROR(AQHBCI_LOGDOMAIN, "Bad crypt head");
      return -1;
    }
    if (m==998) {
      /* PINTAN */
      pinTanMode=1;
      GWEN_KeySpec_SetKeyType(ks, "RSA");
      GWEN_MsgEngine_SetMode(e, "PINTAN");
    }
    else {
      /* DDV/RDH */
      assert(msgkeybuf);
      p=GWEN_DB_GetBinValue(n,
                            "CryptAlgo/MsgKey",
                            0,
                            0,0,
                            &size);
      if (p && size) {
        /* store messsage key */
        if (size<96) {
          DBG_INFO(AQHBCI_LOGDOMAIN, "Filling message key (only %d bytes)",
                   size);
          GWEN_Buffer_FillWithBytes(msgkeybuf, 0, 96-size);
        }
        GWEN_Buffer_AppendBytes(msgkeybuf, p, size);
      }
      else {
        DBG_ERROR(AQHBCI_LOGDOMAIN,
                  "Message key missing (HBCI server error)");
        return -1;
      }
      m=GWEN_DB_GetIntValue(n, "CryptAlgo/keytype", 0, -1);
      if (m==-1) {
        DBG_ERROR(AQHBCI_LOGDOMAIN, "Bad security mode (non-DDV, non-RDH)");
        return -1;
      }
      if (m==6) {
        rdhMode=1;
        GWEN_KeySpec_SetKeyType(ks, "RSA");
        GWEN_MsgEngine_SetMode(e, "RDH");
      }
      else if (m==5) {
        GWEN_KeySpec_SetKeyType(ks, "DES");
        GWEN_MsgEngine_SetMode(e, "DDV");
      }
      else {
        DBG_ERROR(AQHBCI_LOGDOMAIN, "Bad security mode (non-DDV, non-RDH: %d)", m);
        return -1;
      }
    }
  } /* if crypt */
  else {
    int m;

    m=GWEN_DB_GetIntValue(n, "function", 0, -1);
    if (m==-1) {
      DBG_ERROR(AQHBCI_LOGDOMAIN, "Bad sign head");
      return -1;
    }
    if (m==999) {
      /* PINTAN */
      GWEN_MsgEngine_SetMode(e, "PINTAN");
      GWEN_KeySpec_SetKeyType(ks, "RSA");
      pinTanMode=1;
    }
    else {
      assert(signseq);
      *signseq=GWEN_DB_GetIntValue(n, "signseq", 0, 0);
      if (*signseq==0) {
        DBG_ERROR(AQHBCI_LOGDOMAIN, "Invalid signature counter value 0");
        return -1;
      }
      m=GWEN_DB_GetIntValue(n, "SignAlgo/algo", 0, -1);
      if (m==-1) {
        DBG_ERROR(AQHBCI_LOGDOMAIN, "Bad security mode (non-DDV, non-RDH)");
        return -1;
      }
      if (m==10) {
        rdhMode=1;
        GWEN_KeySpec_SetKeyType(ks, "RSA");
        GWEN_MsgEngine_SetMode(e, "RDH");
      }
      else if (m==1) {
        GWEN_KeySpec_SetKeyType(ks, "DES");
        GWEN_MsgEngine_SetMode(e, "DDV");
      }
      else {
        DBG_ERROR(AQHBCI_LOGDOMAIN, "Bad security mode (non-DDV, non-RDH: %d)", m);
        return -1;
      }
    }
  }

  /* store security id */
  if (rdhMode || pinTanMode) {
    /* RDH or PIN/TAN mode */
    s=GWEN_DB_GetCharValue(n,
                           "SecDetails/SecId",
                           0,
                           0);
    if (s) {
      *pSecurityId=s;
      *lSecurityId=strlen(s);
    }
  }
  else {
    /* DDV mode */
    p=GWEN_DB_GetBinValue(n,
                          "SecDetails/SecId",
                          0,
                          0,0,
                          &size);
    if (p) {
      *pSecurityId=p;
      *lSecurityId=size;
    }
  }

  return 0;
}



/* --------------------------------------------------------------- FUNCTION */
int AH_Msg_Decrypt(AH_MSG *hmsg, GWEN_DB_NODE *gr){
  const void *p;
  unsigned int size;
  GWEN_DB_NODE *nhead=0;
  GWEN_DB_NODE *ndata=0;
  GWEN_BUFFER *cdbuf=0;
  GWEN_BUFFER *ndbuf=0;
  GWEN_BUFFER *msgkeybuf=0;
  AH_MEDIUM_RESULT rv;
  GWEN_KEYSPEC *ks=0;
  const char *pSecurityId=0;
  int lSecurityId;
  AH_MEDIUM *medium=0;
  AH_BANK *b;

  assert(hmsg);
  b=AH_Dialog_GetBank(hmsg->dialog);

  /* decrypt */
  nhead=GWEN_DB_GetGroup(gr,
                         GWEN_DB_FLAGS_DEFAULT |
                         GWEN_PATH_FLAGS_NAMEMUSTEXIST,
                         "CryptHead");
  if (!nhead) {
    DBG_ERROR(AQHBCI_LOGDOMAIN, "No crypt head");
    return -1;
  }

  ndata=GWEN_DB_GetGroup(gr,
                         GWEN_DB_FLAGS_DEFAULT |
                         GWEN_PATH_FLAGS_NAMEMUSTEXIST,
                         "CryptData");
  if (!ndata) {
    DBG_ERROR(AQHBCI_LOGDOMAIN, "No crypt data");
    return -1;
  }

  msgkeybuf=GWEN_Buffer_new(0, 128, 0, 1);
  if (AH_Msg_PrepareCryptoSegDec(hmsg,
                                 nhead,
                                 1,
                                 &ks,
                                 0,
                                 &pSecurityId,
                                 &lSecurityId,
                                 msgkeybuf)) {
    DBG_INFO(AQHBCI_LOGDOMAIN, "here");
    GWEN_Buffer_free(msgkeybuf);
    return -1;
  }

  /* get medium */
  DBG_INFO(AQHBCI_LOGDOMAIN, "Getting medium for decrypting");
  medium=AH_HBCI_GetMedium(AH_Bank_GetHbci(b),
                           AH_Dialog_GetDialogOwner(hmsg->dialog),
                           AH_CryptoActionDecrypt);
  if (!medium) {
    DBG_INFO(AQHBCI_LOGDOMAIN,
             "Could not find medium for user \"%s\"",
             GWEN_KeySpec_GetOwner(ks));
    GWEN_Buffer_free(msgkeybuf);
    GWEN_KeySpec_free(ks);
    return -1;
  }

  GWEN_MsgEngine_SetMode(AH_Dialog_GetMsgEngine(hmsg->dialog),
                         AH_Medium_TypeToName(AH_Medium_GetMediumType(medium)));

  p=GWEN_DB_GetBinValue(ndata,
                        "CryptData",
                        0,
                        0,0,
                        &size);
  if (!p) {
    DBG_ERROR(AQHBCI_LOGDOMAIN, "No crypt data");
    GWEN_Buffer_free(msgkeybuf);
    GWEN_KeySpec_free(ks);
    return -1;
  }

  cdbuf=GWEN_Buffer_new((void*)p, size, size, 0);
  GWEN_Buffer_SetMode(cdbuf, 0); /* no dynamic mode ! */

  ndbuf=GWEN_Buffer_new(0, AH_MSG_DEFAULTSIZE, 0, 1);
  GWEN_Buffer_SetStep(ndbuf, AH_MSG_DEFAULTSTEP);

  if (AH_Medium_GetMediumType(medium)==AH_MediumTypePINTAN) {
    GWEN_Buffer_AppendBuffer(ndbuf, cdbuf);
  }
  else {
    rv=AH_Medium_Decrypt(medium,
                         cdbuf,
                         ndbuf,
                         msgkeybuf);
    if (rv!=AH_MediumResultOk) {
      DBG_INFO(AQHBCI_LOGDOMAIN,
               "Error decrypting with medium for user \"%s\"",
               GWEN_KeySpec_GetOwner(ks));
      GWEN_Buffer_free(cdbuf);
      GWEN_Buffer_free(ndbuf);
      GWEN_Buffer_free(msgkeybuf);
      GWEN_KeySpec_free(ks);
      return -1;
    }
  }
  DBG_INFO(AQHBCI_LOGDOMAIN, "Decrypting done");

  /* store new buffer inside message */
  GWEN_Buffer_free(hmsg->origbuffer);
  hmsg->origbuffer=hmsg->buffer;
  GWEN_Buffer_Rewind(ndbuf);
  hmsg->buffer=ndbuf;

  /* store crypter */
  AH_Msg_SetCrypter(hmsg, ks);

  GWEN_Buffer_free(cdbuf);
  GWEN_Buffer_free(msgkeybuf);
  GWEN_KeySpec_free(ks);
  return 0;
}



/* --------------------------------------------------------------- FUNCTION */
int AH_Msg_SequenceCheck(GWEN_DB_NODE *gr) {
  GWEN_DB_NODE *n;
  unsigned int sn;
  unsigned int errors;

  DBG_INFO(AQHBCI_LOGDOMAIN, "Sequence check");

  sn=1;
  errors=0;
  n=GWEN_DB_GetFirstGroup(gr);
  while(n) {
    unsigned int rsn;

    rsn=GWEN_DB_GetIntValue(n, "head/seq", 0, 0);
    if (rsn<900) {
      if (rsn!=sn) {
        DBG_ERROR(AQHBCI_LOGDOMAIN, "Unexpected sequence number (%d, expected %d)",
                  rsn, sn);
        GWEN_DB_Dump(n, stderr, 2);

        GWEN_DB_SetIntValue(n,
                            GWEN_DB_FLAGS_OVERWRITE_VARS,
                            "segment/error/code",
                            9120);
        GWEN_DB_SetCharValue(n,
                             GWEN_DB_FLAGS_OVERWRITE_VARS,
                             "segment/error/text",
                             "Unerwartete Segmentnummer");
        errors++;
      }
      sn++;
    }
    n=GWEN_DB_GetNextGroup(n);
  } /* while */

  if (errors)
    return -1;
  DBG_INFO(AQHBCI_LOGDOMAIN, "Sequence check ok");
  return 0;
}



/* --------------------------------------------------------------- FUNCTION */
int AH_Msg_Verify(AH_MSG *hmsg,
                  GWEN_DB_NODE *gr,
                  unsigned int flags) {
  GWEN_LIST *sigheads;
  GWEN_LIST *sigtails;
  GWEN_DB_NODE *n;
  int nonSigHeads;
  int nSigheads;
  unsigned int dataBegin;
  char *dataStart;
  unsigned int dataLength;
  unsigned int i;
  AH_BANK *b;

  assert(hmsg);
  b=AH_Dialog_GetBank(hmsg->dialog);

  sigheads=GWEN_List_new();

  /* enumerate signature heads */
  nonSigHeads=0;
  nSigheads=0;
  n=GWEN_DB_GetFirstGroup(gr);
  while(n) {
    if (strcasecmp(GWEN_DB_GroupName(n), "SigHead")==0) {
      /* found a signature head */
      if (nonSigHeads) {
        DBG_ERROR(AQHBCI_LOGDOMAIN, "Found some unsigned parts at the beginning");
        GWEN_List_free(sigheads);
        return AB_ERROR_BAD_DATA;
      }
      GWEN_List_PushBack(sigheads, n);
      nSigheads++;
    }
    else if (strcasecmp(GWEN_DB_GroupName(n), "MsgHead")!=0) {
      if (nSigheads)
        break;
      nonSigHeads++;
    }
    n=GWEN_DB_GetNextGroup(n);
  } /* while */

  if (!n) {
    if (nSigheads) {
      DBG_ERROR(AQHBCI_LOGDOMAIN, "Found Signature heads but no other segments");
      GWEN_List_free(sigheads);
      return AB_ERROR_BAD_DATA;
    }
    DBG_INFO(AQHBCI_LOGDOMAIN, "No signatures");
    GWEN_List_free(sigheads);
    return 0;
  }

  /* store begin of signed data */
  dataBegin=GWEN_DB_GetIntValue(n, "segment/pos", 0, 0);
  if (!dataBegin) {
    DBG_ERROR(AQHBCI_LOGDOMAIN, "No position specifications in segment");
    GWEN_List_free(sigheads);
    return AB_ERROR_BAD_DATA;
  }

  /* now get first signature tail */
  while(n) {
    if (strcasecmp(GWEN_DB_GroupName(n), "SigTail")==0) {
      unsigned int currpos;

      /* found a signature tail */
      currpos=GWEN_DB_GetIntValue(n, "segment/pos", 0, 0);
      if (!currpos || dataBegin>currpos) {
        DBG_ERROR(AQHBCI_LOGDOMAIN, "Bad position specification in Signature tail");
        GWEN_List_free(sigheads);
        return AB_ERROR_BAD_DATA;
      }
      dataLength=currpos-dataBegin;
      break;
    }
    n=GWEN_DB_GetNextGroup(n);
  } /* while */

  if (!n) {
    DBG_ERROR(AQHBCI_LOGDOMAIN, "No signature tail found");
    GWEN_List_free(sigheads);
    return AB_ERROR_BAD_DATA;
  }

  sigtails=GWEN_List_new();
  while(n) {
    if (strcasecmp(GWEN_DB_GroupName(n), "SigTail")!=0)
      break;
    GWEN_List_PushBack(sigtails, n);
    n=GWEN_DB_GetNextGroup(n);
  } /* while */

  if (!n) {
    DBG_ERROR(AQHBCI_LOGDOMAIN, "Message tail expected");
    GWEN_List_free(sigheads);
    GWEN_List_free(sigtails);
    return AB_ERROR_BAD_DATA;
  }

  if (strcasecmp(GWEN_DB_GroupName(n), "MsgTail")!=0) {
    DBG_ERROR(AQHBCI_LOGDOMAIN, "Unexpected segment (msg tail expected)");
    GWEN_List_free(sigheads);
    GWEN_List_free(sigtails);
    return AB_ERROR_BAD_DATA;
  }

  n=GWEN_DB_GetNextGroup(n);
  if (n) {
    DBG_ERROR(AQHBCI_LOGDOMAIN, "Unexpected segment (end expected)");
    GWEN_List_free(sigheads);
    GWEN_List_free(sigtails);
    return AB_ERROR_BAD_DATA;
  }

  if (GWEN_List_GetSize(sigheads)!=
      GWEN_List_GetSize(sigtails)) {
    DBG_ERROR(AQHBCI_LOGDOMAIN,
              "Number of signature heads (%d) does not match "
              "number of signature tails (%d)",
              GWEN_List_GetSize(sigheads),
              GWEN_List_GetSize(sigtails));
    GWEN_List_free(sigheads);
    GWEN_List_free(sigtails);
    return AB_ERROR_BAD_DATA;
  }

  /* ok, now verify all signatures */
  dataStart=GWEN_Buffer_GetStart(hmsg->buffer)+dataBegin;
  for (i=0; i< GWEN_List_GetSize(sigtails); i++) {
    GWEN_DB_NODE *sighead;
    GWEN_DB_NODE *sigtail;
    GWEN_BUFFER *dbuf;
    GWEN_BUFFER *sigbuf=0;
    const void *p;
    unsigned int size;
    GWEN_KEYSPEC *ks=0;
    int signSeq;
    const char *pSecurityId=0;
    int lSecurityId;
    AH_MEDIUM *medium=0;

    /* get signature tail */
    sigtail=(GWEN_DB_NODE*)GWEN_List_GetBack(sigtails);

    /* get corresponding signature head */
    sighead=(GWEN_DB_NODE*)GWEN_List_GetFront(sigheads);

    if (!sighead || !sigtail) {
      DBG_ERROR(AQHBCI_LOGDOMAIN, "No signature head/tail left (internal error)");
      GWEN_List_free(sigheads);
      GWEN_List_free(sigtails);
      return AB_ERROR_BAD_DATA;
    }

    GWEN_List_PopBack(sigtails);
    GWEN_List_PopFront(sigheads);

    /* some checks */
    if (strcasecmp(GWEN_DB_GetCharValue(sighead, "ctrlref", 0, ""),
                   GWEN_DB_GetCharValue(sigtail, "ctrlref", 0, ""))!=0) {
      DBG_ERROR(AQHBCI_LOGDOMAIN, "Non-matching signature tail");
      GWEN_List_free(sigheads);
      GWEN_List_free(sigtails);
      return AB_ERROR_BAD_DATA;
    }

    /* prepare data buffer */
    dbuf=GWEN_Buffer_new(0,
                         dataLength+GWEN_DB_GetIntValue(sighead,
                                                        "segment/length",
                                                        0,
                                                        0),
                         0,1);
    GWEN_Buffer_AppendBytes(dbuf,
                            GWEN_Buffer_GetStart(hmsg->buffer)+
                            GWEN_DB_GetIntValue(sighead,
                                                "segment/pos",
                                                0,
                                                0),
                            GWEN_DB_GetIntValue(sighead,
                                                "segment/length",
                                                0,
                                                0));
    GWEN_Buffer_AppendBytes(dbuf, dataStart, dataLength);

    /* prepare signature buffer */
    p=GWEN_DB_GetBinValue(sigtail, "signature", 0, 0, 0, &size);
    if (!p) {
      DBG_DEBUG(AQHBCI_LOGDOMAIN, "No signature");
    }

    GWEN_Buffer_Rewind(dbuf);

    if (p && size) {
      sigbuf=GWEN_Buffer_new((char*)p, size, size, 0);
      GWEN_Buffer_SetMode(sigbuf, 0);
    }

    if (AH_Msg_PrepareCryptoSegDec(hmsg,
                                   sighead,
                                   0,
                                   &ks,
                                   &signSeq,
                                   &pSecurityId,
                                   &lSecurityId,
                                   0)) {
      GWEN_Buffer_free(sigbuf);
      GWEN_Buffer_free(dbuf);
      GWEN_List_free(sigheads);
      GWEN_List_free(sigtails);
      GWEN_KeySpec_free(ks);
      DBG_INFO(AQHBCI_LOGDOMAIN, "here");
      return AB_ERROR_BAD_DATA;
    }

    DBG_NOTICE(AQHBCI_LOGDOMAIN, "Got sign seq %d", signSeq);

    /* get medium */
    DBG_INFO(AQHBCI_LOGDOMAIN, "Getting medium for verification");
    medium=AH_HBCI_GetMedium(AH_Bank_GetHbci(b),
                             AH_Dialog_GetDialogOwner(hmsg->dialog),
                             AH_CryptoActionVerify);
    if (!medium) {
      char *nb;
      GWEN_KEYSPEC *nks;

      /* medium not found */
      nks=GWEN_KeySpec_dup(ks);
      p=GWEN_KeySpec_GetOwner(nks);
      assert(p);
      nb=(char*)malloc(strlen(p)+2);
      assert(nb);
      nb[0]='?';
      nb[1]=0;
      strcat(nb, p);
      GWEN_KeySpec_SetOwner(nks, nb);
      AH_Msg_AddSigner(hmsg, nks);
      GWEN_KeySpec_free(nks);
      free(nb);
      DBG_WARN(AQHBCI_LOGDOMAIN,
               "Unknown signer \"%s\", no verification",
               GWEN_KeySpec_GetOwner(ks));
      GWEN_DB_SetIntValue(sighead,
                          GWEN_DB_FLAGS_OVERWRITE_VARS,
                          "segment/error/code",
                          3300);
      GWEN_DB_SetCharValue(sighead,
                           GWEN_DB_FLAGS_OVERWRITE_VARS,
                           "segment/error/text",
                           "Unbekannter Signierer");
    } /* if !medium */
    else {
      AH_MEDIUM_RESULT mres;

      /* verify signature */
      if (AH_Medium_GetMediumType(medium)==AH_MediumTypePINTAN) {
#if 0
        const char *msgPin;
        const char *msgTan;
#endif

        /* PIN/TAN mode */
        GWEN_Buffer_free(sigbuf);
        GWEN_Buffer_free(dbuf);

#if 0 /* disabled, since we have no PIN or TAN from the server to verify */
        /* check PIN and TAN */
        msgPin=GWEN_DB_GetCharValue(sigtail, "pin", 0, 0);
        if (!msgPin) {
          DBG_INFO(AQHBCI_LOGDOMAIN, "No PIN to verify");
          GWEN_DB_SetIntValue(sighead,
                              GWEN_DB_FLAGS_OVERWRITE_VARS,
                              "segment/error/code",
                              901);
          GWEN_DB_SetCharValue(sighead,
                               GWEN_DB_FLAGS_OVERWRITE_VARS,
                               "segment/error/text",
                               "PIN fehlt, wird aber auch nicht erwartet");
        }
        else {
          /* we have a PIN, check it */
          char mediumPin[7];

          if (AH_MediumPINTAN_GetPin(medium, mediumPin,
                                     0, sizeof(mediumPin),
                                     0)) {
            char *nb;
            GWEN_KEYSPEC *nks;
  
            nks=GWEN_KeySpec_dup(ks);
            p=GWEN_KeySpec_GetOwner(nks);
            assert(p);
            nb=(char*)malloc(strlen(p)+2);
            assert(nb);
            nb[0]='!';
            nb[1]=0;
            strcat(nb, p);
            GWEN_KeySpec_SetOwner(nks, nb);
            AH_Msg_AddSigner(hmsg, nks);
            GWEN_KeySpec_free(nks);
            free(nb);
            DBG_NOTICE(AQHBCI_LOGDOMAIN, "Error getting PIN of medium");
            GWEN_DB_SetIntValue(sighead,
                                GWEN_DB_FLAGS_OVERWRITE_VARS,
                                "segment/error/code",
                                9400);
            GWEN_DB_SetCharValue(sighead,
                                 GWEN_DB_FLAGS_OVERWRITE_VARS,
                                 "segment/error/text",
                                 "Interner Fehler des Mediums");
          }
          else {
            if (strcasecmp(msgPin, mediumPin)!=0) {
              char *nb;
              GWEN_KEYSPEC *nks;
    
              nks=GWEN_KeySpec_dup(ks);
              p=GWEN_KeySpec_GetOwner(nks);
              assert(p);
              nb=(char*)malloc(strlen(p)+2);
              assert(nb);
              nb[0]='!';
              nb[1]=0;
              strcat(nb, p);
              GWEN_KeySpec_SetOwner(nks, nb);
              AH_Msg_AddSigner(hmsg, nks);
              GWEN_KeySpec_free(nks);
              free(nb);
              DBG_NOTICE(AQHBCI_LOGDOMAIN, "Invalid PIN");
              GWEN_DB_SetIntValue(sighead,
                                  GWEN_DB_FLAGS_OVERWRITE_VARS,
                                  "segment/error/code",
                                  9942);
              GWEN_DB_SetCharValue(sighead,
                                   GWEN_DB_FLAGS_OVERWRITE_VARS,
                                   "segment/error/text",
                                   "PIN ungueltig");
            }
            else {
              GWEN_DB_SetIntValue(sighead,
                                  GWEN_DB_FLAGS_OVERWRITE_VARS,
                                  "segment/error/code",
                                  901);
              GWEN_DB_SetCharValue(sighead,
                                   GWEN_DB_FLAGS_OVERWRITE_VARS,
                                   "segment/error/text",
                                   "PIN gueltig");
              AH_Msg_AddSigner(hmsg, ks);
              DBG_INFO(AQHBCI_LOGDOMAIN, "Signature is valid");
            }
          }
        }

        /* get TAN, if any */
        msgTan=GWEN_DB_GetCharValue(sigtail, "tan", 0, 0);
        if (msgTan) {
          /* FIXME: later check TAN with medium instead */
          AH_Msg_SetTan(hmsg, msgTan);
        }
#endif
      } /* if PIN/TAN */
      else {
        /* DDV or RDH mode */
        GWEN_Buffer_Rewind(dbuf);

        if (GWEN_Buffer_GetUsedBytes(sigbuf)==0) {
          DBG_ERROR(AQHBCI_LOGDOMAIN, "No signature");
          GWEN_List_free(sigheads);
          GWEN_List_free(sigtails);
          GWEN_KeySpec_free(ks);
          return AB_ERROR_BAD_DATA;
        }

        DBG_NOTICE(AQHBCI_LOGDOMAIN, "Checking with sign seq %d", signSeq);
        /* TODO: Extend signature counter width and really check it */
        /* mres=AH_Medium_Verify(medium, dbuf, sigbuf, signSeq); */
        mres=AH_Medium_Verify(medium, dbuf, sigbuf, 0);
        GWEN_Buffer_free(sigbuf);
        GWEN_Buffer_free(dbuf);

        if (mres==AH_MediumResultOk) {
          /* signature ok, add signer */
          AH_Msg_AddSigner(hmsg, ks);
          DBG_INFO(AQHBCI_LOGDOMAIN, "Signature is valid");
          GWEN_DB_SetIntValue(sighead,
                              GWEN_DB_FLAGS_OVERWRITE_VARS,
                              "segment/error/code",
                              10);
          GWEN_DB_SetCharValue(sighead,
                               GWEN_DB_FLAGS_OVERWRITE_VARS,
                               "segment/error/text",
                               "Signierer bekannt");

          GWEN_DB_SetIntValue(sigtail,
                              GWEN_DB_FLAGS_OVERWRITE_VARS,
                              "segment/error/code",
                              10);
          GWEN_DB_SetCharValue(sigtail,
                               GWEN_DB_FLAGS_OVERWRITE_VARS,
                               "segment/error/text",
                               "Signatur ist gueltig");
        }
        else if (mres==AH_MediumResultNoKey) {
          char *nb;
          GWEN_KEYSPEC *nks;

          DBG_WARN(AQHBCI_LOGDOMAIN, "No key to verify signature");

          nks=GWEN_KeySpec_dup(ks);
          p=GWEN_KeySpec_GetOwner(nks);
          assert(p);
          nb=(char*)malloc(strlen(p)+2);
          assert(nb);
          nb[0]='~';
          nb[1]=0;
          strcat(nb, p);
          GWEN_KeySpec_SetOwner(nks, nb);
          AH_Msg_AddSigner(hmsg, nks);
          GWEN_KeySpec_free(nks);
          free(nb);

          GWEN_DB_SetIntValue(sigtail,
                              GWEN_DB_FLAGS_OVERWRITE_VARS,
                              "segment/error/code",
                              3340);
          GWEN_DB_SetCharValue(sigtail,
                               GWEN_DB_FLAGS_OVERWRITE_VARS,
                               "segment/error/text",
                               "Unbekannter Signierer");
        }
        else {
          char *nb;
          GWEN_KEYSPEC *nks;

          if (mres==AH_MediumResultSignSeq) {
            DBG_WARN(AQHBCI_LOGDOMAIN, "Double use of signature");
            GWEN_DB_SetIntValue(sigtail,
                                GWEN_DB_FLAGS_OVERWRITE_VARS,
                                "segment/error/code",
                                9390);
            GWEN_DB_SetCharValue(sigtail,
                                 GWEN_DB_FLAGS_OVERWRITE_VARS,
                                 "segment/error/text",
                                 "Doppeleinreichung");
          }
          else if (mres==AH_MediumResultInvalidSignature) {
            DBG_WARN(AQHBCI_LOGDOMAIN, "Invalid signature");
            GWEN_DB_SetIntValue(sigtail,
                                GWEN_DB_FLAGS_OVERWRITE_VARS,
                                "segment/error/code",
                                9340);
            GWEN_DB_SetCharValue(sigtail,
                                 GWEN_DB_FLAGS_OVERWRITE_VARS,
                                 "segment/error/text",
                                 "Elektronische Signatur falsch");
          }
          else {
            DBG_WARN(AQHBCI_LOGDOMAIN, "Error in security medium");
            GWEN_DB_SetIntValue(sigtail,
                                GWEN_DB_FLAGS_OVERWRITE_VARS,
                                "segment/error/code",
                                9400);
            GWEN_DB_SetCharValue(sigtail,
                                 GWEN_DB_FLAGS_OVERWRITE_VARS,
                                 "segment/error/text",
                                 "Allgemeiner Fehler des Sicherheitsmediums");
          }

          nks=GWEN_KeySpec_dup(ks);
          p=GWEN_KeySpec_GetOwner(nks);
          assert(p);
          nb=(char*)malloc(strlen(p)+2);
          assert(nb);
          nb[0]='!';
          nb[1]=0;
          strcat(nb, p);
          GWEN_KeySpec_SetOwner(nks, nb);
          AH_Msg_AddSigner(hmsg, nks);
          GWEN_KeySpec_free(nks);
          free(nb);
        }
      } /* if we have a medium */
    } /* if not PIN/TAN */
    GWEN_KeySpec_free(ks);
    DBG_INFO(AQHBCI_LOGDOMAIN, "Verification done");
  } /* for */

  GWEN_List_free(sigheads);
  GWEN_List_free(sigtails);
  return 0;
}



/* --------------------------------------------------------------- FUNCTION */
int AH_Msg_DecodeMsg(AH_MSG *hmsg,
                     GWEN_DB_NODE *gr,
                     unsigned int flags) {
  GWEN_MSGENGINE *e;
  int rv;
  GWEN_DB_NODE *n, *n2;
  AH_CUSTOMER *cu;
  AH_USER *u;
  AH_MEDIUM *m;
  const char *mode;
  GWEN_TYPE_UINT32 expMsgNum;

  e=AH_Dialog_GetMsgEngine(hmsg->dialog);
  assert(e);

  /* set mode */
  cu=AH_Dialog_GetDialogOwner(hmsg->dialog);
  assert(cu);
  u=AH_Customer_GetUser(cu);
  assert(u);
  m=AH_User_GetMedium(u);
  assert(m);
  mode=AH_Medium_TypeToName(AH_Medium_GetMediumType(m));
  DBG_NOTICE(AQHBCI_LOGDOMAIN, "Mode is: %s", mode);
  GWEN_MsgEngine_SetMode(e, mode);

  DBG_NOTICE(AQHBCI_LOGDOMAIN, "Decoding message:");

  /* log encrypted message */
  AH_Msg_LogMessage(hmsg, hmsg->buffer, 1, 1);

  GWEN_Buffer_Rewind(hmsg->buffer);
  rv=AH_Msg_ReadMessage(hmsg, e, "SEG", hmsg->buffer, gr, flags);
  if (rv) {
    DBG_INFO(AQHBCI_LOGDOMAIN, "here");
    return AB_ERROR_BAD_DATA;
  }

  /* take some values out of the message head (with checks) */
  n=GWEN_DB_GetGroup(gr,
                     GWEN_DB_FLAGS_DEFAULT |
                     GWEN_PATH_FLAGS_NAMEMUSTEXIST,
                     "MsgHead");
  if (!n) {
    DBG_ERROR(AQHBCI_LOGDOMAIN, "No message head segment");
    return AB_ERROR_BAD_DATA;
  }

  /* sample message number */
  expMsgNum=AH_Dialog_GetLastMsgNum(hmsg->dialog);
  hmsg->msgNum=GWEN_DB_GetIntValue(n, "msgnum", 0, 0);
  if (AH_Dialog_CheckReceivedMsgNum(hmsg->dialog, hmsg->msgNum)) {
    DBG_WARN(AQHBCI_LOGDOMAIN, "Bad message number, ignoring");
    AB_Banking_ProgressLog(AH_Dialog_GetBankingApi(hmsg->dialog),
			   0,
			   AB_Banking_LogLevelWarn,
                           I18N("Bad message number, ignoring"));
    /* FIXME: this is sometimes returned by the Sparkasse server, when
     * the server is currently out of service
     * We will for now ignore this, since this message still returns an
     * error code and text which might be of interest to the user.
     */
    /* return -1; */
    hmsg->msgNum=expMsgNum;
  }

  /* sample dialog id if necessary */
  if (strcasecmp(AH_Dialog_GetDialogId(hmsg->dialog), "0")==0) {
    /* dialog id not yet known, copy it */
    const char *p;

    p=GWEN_DB_GetCharValue(n, "dialogid", 0, 0);
    if (!p) {
      DBG_ERROR(AQHBCI_LOGDOMAIN, "No dialog id in message");
      return AB_ERROR_BAD_DATA;
    }
    if (AH_Dialog_GetFlags(hmsg->dialog) & AH_DIALOG_FLAGS_INITIATOR) {
      DBG_INFO(AQHBCI_LOGDOMAIN, "Setting new dialog id (%s)", p);
      AH_Dialog_SetDialogId(hmsg->dialog, p);
      GWEN_MsgEngine_SetValue(e, "dialogid", p);
    }
    else {
      const char *myDlgId;

      myDlgId=AH_Dialog_GetDialogId(hmsg->dialog);
      assert(myDlgId);
      if (strcasecmp(myDlgId, p)!=0) {
        DBG_ERROR(AQHBCI_LOGDOMAIN, "Bad dialog id received from peer");
        GWEN_DB_SetIntValue(n, GWEN_DB_FLAGS_OVERWRITE_VARS,
                            "segment/error/code", 9210);
        GWEN_DB_SetCharValue(n, GWEN_DB_FLAGS_OVERWRITE_VARS,
                             "segment/error/text",
                             "Ungueltige Dialog Kennung");
        return AB_ERROR_BAD_DATA;
      } /* if bad dialog id */
    } /* if !initialtor */
  }
  n2=GWEN_DB_GetGroup(n,
                      GWEN_DB_FLAGS_DEFAULT |
                      GWEN_PATH_FLAGS_NAMEMUSTEXIST,
                      "msgref");
  if (n2) {
    /* we have a message reference, get it */
    const char *p;

    DBG_INFO(AQHBCI_LOGDOMAIN, "Found a message reference");
    p=GWEN_DB_GetCharValue(n2, "dialogid", 0, 0);
    if (!p) {
      DBG_ERROR(AQHBCI_LOGDOMAIN, "No reference dialog id in message");
      return AB_ERROR_BAD_DATA;
    }
    if (strcasecmp(AH_Dialog_GetDialogId(hmsg->dialog), p)!=0) {
      DBG_ERROR(AQHBCI_LOGDOMAIN, "Dialog id does not match current dialog id !");
      return AB_ERROR_BAD_DATA;
    }
    hmsg->refMsgNum=GWEN_DB_GetIntValue(n2, "msgnum", 0, 0);
    if (hmsg->refMsgNum!=expMsgNum) {
      DBG_WARN(AQHBCI_LOGDOMAIN, "Bad message reference number, ignoring");
      AB_Banking_ProgressLog(AH_Dialog_GetBankingApi(hmsg->dialog),
			     0,
			     AB_Banking_LogLevelWarn,
                             I18N("Bad message reference number, ignoring"));
      hmsg->refMsgNum=expMsgNum;
    }
  }
  else {
    DBG_INFO(AQHBCI_LOGDOMAIN, "No message reference found");
  }

  /* find Crypt head */
  n=GWEN_DB_GetGroup(gr,
                     GWEN_DB_FLAGS_DEFAULT |
                     GWEN_PATH_FLAGS_NAMEMUSTEXIST,
                     "CryptHead");
  if (n) {
    if (GWEN_DB_GetIntValue(n, "segment/error/code", 0, 0)>=9000) {
      DBG_ERROR(AQHBCI_LOGDOMAIN, "Encryption error");
      return AB_ERROR_GENERIC;
    }
    rv=AH_Msg_Decrypt(hmsg, gr);
    if (rv) {
      DBG_INFO(AQHBCI_LOGDOMAIN, "here");
      return AB_ERROR_SECURITY;
    }
    /* unlink and delete crypthead */
    GWEN_DB_UnlinkGroup(n);
    GWEN_DB_Group_free(n);

    /* unlink and delete cryptdata */
    n=GWEN_DB_GetGroup(gr,
                       GWEN_DB_FLAGS_DEFAULT |
                       GWEN_PATH_FLAGS_NAMEMUSTEXIST,
                       "CryptData");
    if (n) {
      GWEN_DB_UnlinkGroup(n);
      GWEN_DB_Group_free(n);
    }

    /* parse decrypted message part */
    n=GWEN_DB_GetGroup(gr,
                       GWEN_DB_FLAGS_DEFAULT |
                       GWEN_PATH_FLAGS_NAMEMUSTEXIST,
                       "MsgTail");
    if (n) {
      /* temporarily unlink MsgTail, it will be appended after decoding
       * the crypted part, to keep the segment sequence correct */
      GWEN_DB_UnlinkGroup(n);
    }
    rv=AH_Msg_ReadMessage(hmsg, e, "SEG", hmsg->buffer, gr, flags);
    if (n)
      GWEN_DB_AddGroup(gr, n);
    if (rv) {
      DBG_INFO(AQHBCI_LOGDOMAIN, "here");
      return AB_ERROR_BAD_DATA;
    }
  } /* if crypthead */

  /* check segment sequence numbers */
  rv=AH_Msg_SequenceCheck(gr);
  if (rv) {
    DBG_INFO(AQHBCI_LOGDOMAIN, "here");
    return AB_ERROR_SECURITY;
  }

  /* verify signatures */
  rv=AH_Msg_Verify(hmsg, gr, flags);
  if (rv) {
    DBG_INFO(AQHBCI_LOGDOMAIN, "here");
    return rv;
  }

  /* log decrypted message */
  AH_Msg_LogMessage(hmsg, hmsg->buffer, 1, 0);

  return 0;
}



/* --------------------------------------------------------------- FUNCTION */
void AH_Msg__Dump(const AH_MSG *hmsg,
                  FILE *f,
                  unsigned int indent) {
  unsigned int i;
  const GWEN_KEYSPEC *ks;

  for (i=0; i<indent; i++) fprintf(f, " ");
  fprintf(f, "AH_Msg\n");
  for (i=0; i<indent; i++) fprintf(f, " ");
  fprintf(f, "==================================================\n");
  for (i=0; i<indent; i++) fprintf(f, " ");
  if (hmsg->origbuffer) {
    for (i=0; i<indent; i++) fprintf(f, " ");
    fprintf(f, "Original buffer      :\n");
    GWEN_Buffer_Dump(hmsg->origbuffer, f, indent+2);
  }
  else {
    for (i=0; i<indent; i++) fprintf(f, " ");
    fprintf(f, "Original buffer      : none\n");
  }
  if (hmsg->buffer) {
    for (i=0; i<indent; i++) fprintf(f, " ");
    fprintf(f, "Buffer:\n");
    GWEN_Buffer_Dump(hmsg->buffer, f, indent+2);
  }
  else {
    for (i=0; i<indent; i++) fprintf(f, " ");
    fprintf(f, "Buffer               : none\n");
  }

  for (i=0; i<indent; i++) fprintf(f, " ");
  if (hmsg->crypter) {
    fprintf(f, "Crypter:\n");
    GWEN_KeySpec_Dump(hmsg->crypter, f, indent+2);
  }
  else {
    fprintf(f, "Crypter: none\n");
  }
  for (i=0; i<indent; i++) fprintf(f, " ");
  fprintf(f, "Signers (%d):\n", hmsg->nSigners);
  ks=hmsg->signers;
  while(ks) {
    GWEN_KeySpec_Dump(ks, f, indent+2);
    ks=GWEN_KeySpec_ConstNext(ks);
  } /* while */
  for (i=0; i<indent; i++) fprintf(f, " ");
  fprintf(f, "Nodes                : %d\n", hmsg->nodes);
  for (i=0; i<indent; i++) fprintf(f, " ");
  fprintf(f, "Msg number           : %d\n", hmsg->msgNum);
  for (i=0; i<indent; i++) fprintf(f, " ");
  fprintf(f, "Reference msg number : %d\n", hmsg->refMsgNum);
  for (i=0; i<indent; i++) fprintf(f, " ");
  fprintf(f, "First segment        : %d\n", hmsg->firstSegment);
  for (i=0; i<indent; i++) fprintf(f, " ");
  fprintf(f, "Last segment         : %d\n", hmsg->lastSegment);
  for (i=0; i<indent; i++) fprintf(f, " ");
  fprintf(f, "\n");
}



/* --------------------------------------------------------------- FUNCTION */
AH_DIALOG *AH_Msg_GetDialog(const AH_MSG *hmsg){
  assert(hmsg);
  return hmsg->dialog;
}



/* --------------------------------------------------------------- FUNCTION */
int AH_Msg_IsSignedBy(const AH_MSG *hmsg, const char *s){
  GWEN_KEYSPEC *ks;

  ks=AH_Msg_GetSigners(hmsg);
  while(ks) {
    if (strcasecmp(GWEN_KeySpec_GetOwner(ks), s)==0)
      break;
    ks=GWEN_KeySpec_Next(ks);
  } /* while */
  if (ks) {
    DBG_INFO(AQHBCI_LOGDOMAIN, "Message signed by \"%s\"", s);
    return 1;
  }
  else {
    DBG_INFO(AQHBCI_LOGDOMAIN, "Customer \"%s\" did not sign the message", s);
    return 0;
  }
}



/* --------------------------------------------------------------- FUNCTION */
int AH_Msg_EnableInsert(AH_MSG *hmsg) {
  assert(hmsg);

  if (hmsg->nodes && !hmsg->enableInsert) {
    DBG_INFO(AQHBCI_LOGDOMAIN, "There already are nodes in the buffer, "
             "cannot insert");
    return AB_ERROR_INVALID;
  }
  hmsg->enableInsert=1;
  return 0;
}



/* --------------------------------------------------------------- FUNCTION */
int AH_Msg_HasWarnings(const AH_MSG *hmsg) {
  assert(hmsg);
  return hmsg->hasWarnings;
}



/* --------------------------------------------------------------- FUNCTION */
void AH_Msg_SetHasWarnings(AH_MSG *hmsg, int i) {
  assert(hmsg);
  hmsg->hasWarnings=i;
}



/* --------------------------------------------------------------- FUNCTION */
int AH_Msg_HasErrors(const AH_MSG *hmsg) {
  assert(hmsg);
  return hmsg->hasErrors;
}



/* --------------------------------------------------------------- FUNCTION */
void AH_Msg_SetHasErrors(AH_MSG *hmsg, int i) {
  assert(hmsg);
  hmsg->hasErrors=i;
}



/* --------------------------------------------------------------- FUNCTION */
int AH_Msg_GetResultCode(const AH_MSG *hmsg){
  assert(hmsg);
  return hmsg->resultCode;
}



/* --------------------------------------------------------------- FUNCTION */
void AH_Msg_SetResultCode(AH_MSG *hmsg, int i){
  assert(hmsg);
  hmsg->resultCode=i;
}



/* --------------------------------------------------------------- FUNCTION */
const char *AH_Msg_GetResultText(const AH_MSG *hmsg){
  assert(hmsg);
  return hmsg->resultText;
}



/* --------------------------------------------------------------- FUNCTION */
void AH_Msg_SetResultText(AH_MSG *hmsg, const char *s){
  assert(hmsg);
  free(hmsg->resultText);
  if (s) hmsg->resultText=strdup(s);
  else hmsg->resultText=0;
}



/* --------------------------------------------------------------- FUNCTION */
const char *AH_Msg_GetResultParam(const AH_MSG *hmsg){
  assert(hmsg);
  return hmsg->resultParam;
}



/* --------------------------------------------------------------- FUNCTION */
void AH_Msg_SetResultParam(AH_MSG *hmsg, const char *s){
  assert(hmsg);
  free(hmsg->resultParam);
  if (s) hmsg->resultParam=strdup(s);
  else hmsg->resultParam=0;
}



/* --------------------------------------------------------------- FUNCTION */
unsigned int AH_Msg_GetHbciVersion(const AH_MSG *hmsg) {
  assert(hmsg);
  return hmsg->hbciVersion;
}



/* --------------------------------------------------------------- FUNCTION */
void AH_Msg_SetHbciVersion(AH_MSG *hmsg, unsigned int i) {
  assert(hmsg);
  hmsg->hbciVersion=i;
}



/* --------------------------------------------------------------- FUNCTION */
void AH_Msg_SetDecodedMsg(AH_MSG *msg, GWEN_DB_NODE *db){
  assert(msg);
  assert(db);

  GWEN_DB_Group_free(msg->decodedMsg);
  msg->decodedMsg=db;
}



/* --------------------------------------------------------------- FUNCTION */
GWEN_DB_NODE *AH_Msg_GetDecodedMsg(const AH_MSG *msg){
  assert(msg);
  return msg->decodedMsg;
}



/* --------------------------------------------------------------- FUNCTION */
GWEN_DB_NODE *AH_Msg_TakeDecodedMsg(AH_MSG *msg){
  GWEN_DB_NODE *db;

  assert(msg);
  db=msg->decodedMsg;
  msg->decodedMsg=0;
  return db;
}



/* --------------------------------------------------------------- FUNCTION */
void AH_Msg_SetTan(AH_MSG *hmsg, const char *s){
  assert(hmsg);
  free(hmsg->usedTan);
  if (s) hmsg->usedTan=strdup(s);
  else hmsg->usedTan=0;
}



/* --------------------------------------------------------------- FUNCTION */
const char *AH_Msg_GetTan(const AH_MSG *hmsg){
  assert(hmsg);
  return hmsg->usedTan;
}



/* --------------------------------------------------------------- FUNCTION */
int AH_Msg_GetNeedTan(const AH_MSG *hmsg){
  assert(hmsg);
  return hmsg->needTan;
}



/* --------------------------------------------------------------- FUNCTION */
void AH_Msg_SetNeedTan(AH_MSG *hmsg, int i){
  assert(hmsg);
  hmsg->needTan=i;
}



/* --------------------------------------------------------------- FUNCTION */
int AH_Msg_NoSysId(const AH_MSG *hmsg){
  assert(hmsg);
  return hmsg->noSysId;
}



/* --------------------------------------------------------------- FUNCTION */
void AH_Msg_SetNoSysId(AH_MSG *hmsg, int i){
  assert(hmsg);
  hmsg->noSysId=i;
}



/* --------------------------------------------------------------- FUNCTION */
void AH_Msg_LogMessage(AH_MSG *msg,
                       GWEN_BUFFER *buf,
                       int rec,
                       int crypt) {
  GWEN_DB_NODE *db;
  AH_CUSTOMER *cu;
  AH_USER *u;
  AH_MEDIUM *m;
  AH_BANK *b;
  AH_HBCI *h;
  GWEN_BUFFEREDIO *bio;
  int fd;
  GWEN_ERRORCODE err;
  unsigned int bsize;
  const char *logFile;
  int vmajor, vminor, vpatchlevel, vbuild;
  char vbuf[32];

  assert(msg);
  assert(buf);

  logFile=AH_Dialog_GetLogFile(msg->dialog);
  if (!logFile) {
    DBG_ERROR(AQHBCI_LOGDOMAIN,
              "No log file, logging disabled for this dialog");
    return;
  }

  db=GWEN_DB_Group_new("header");
  cu=AH_Dialog_GetDialogOwner(msg->dialog);
  u=AH_Customer_GetUser(cu);
  b=AH_User_GetBank(u);
  m=AH_User_GetMedium(u);
  h=AH_Medium_GetHBCI(m);

  GWEN_DB_SetCharValue(db, GWEN_DB_FLAGS_OVERWRITE_VARS,
                       "mode",
                       AH_Medium_TypeToName(AH_Medium_GetMediumType(m)));
  GWEN_DB_SetCharValue(db, GWEN_DB_FLAGS_OVERWRITE_VARS,
                       "crypt",
                       crypt?"yes":"no");
  GWEN_DB_SetCharValue(db, GWEN_DB_FLAGS_OVERWRITE_VARS,
                       "sender",
                       rec?"bank":"user");
  GWEN_Version(&vmajor, &vminor, &vpatchlevel, &vbuild);
  snprintf(vbuf, sizeof(vbuf), "%d.%d.%d.%d", vmajor, vminor, vpatchlevel, vbuild);
  GWEN_DB_SetCharValue(db, GWEN_DB_FLAGS_OVERWRITE_VARS,
                       "gwenhywfar", vbuf);
  GWEN_DB_SetCharValue(db, GWEN_DB_FLAGS_OVERWRITE_VARS,
                       "aqhbci",
                       AQHBCI_VERSION_FULL_STRING);
  GWEN_DB_SetCharValue(db, GWEN_DB_FLAGS_OVERWRITE_VARS,
                       "appname",
                       AH_HBCI_GetProductName(h));
  GWEN_DB_SetCharValue(db, GWEN_DB_FLAGS_OVERWRITE_VARS,
                       "appversion",
                       AH_HBCI_GetProductVersion(h));
  GWEN_DB_SetIntValue(db, GWEN_DB_FLAGS_OVERWRITE_VARS,
                      "size",
                      GWEN_Buffer_GetUsedBytes(buf));
  if (GWEN_Directory_GetPath(logFile, GWEN_PATH_FLAGS_VARIABLE)) {
    DBG_ERROR(AQHBCI_LOGDOMAIN,
              "Path \"%s\" is not available, cannot log",
              logFile);
    GWEN_DB_Group_free(db);
    return;
  }
  fd=open(logFile,
          O_RDWR | O_CREAT | O_APPEND);
  if (fd==-1) {
    DBG_ERROR(AQHBCI_LOGDOMAIN,
              "open(%s): %s",
              logFile, strerror(errno));
    GWEN_DB_Group_free(db);
    return;
  }

  bio=GWEN_BufferedIO_File_new(fd);
  GWEN_BufferedIO_SetWriteBuffer(bio, 0, 1024);
  if (GWEN_DB_WriteToStream(db, bio,
                            GWEN_DB_FLAGS_HTTP |
                            GWEN_DB_FLAGS_OMIT_TYPES)) {
    DBG_ERROR(AQHBCI_LOGDOMAIN,
              "Could write log header to \"%s\", cannot log",
              logFile);
    GWEN_BufferedIO_Abandon(bio);
    GWEN_BufferedIO_free(bio);
    GWEN_DB_Group_free(db);
    return;
  }

  err=GWEN_BufferedIO_WriteLine(bio, "");
  if (!GWEN_Error_IsOk(err)) {
    DBG_ERROR_ERR(AQHBCI_LOGDOMAIN, err);
    GWEN_BufferedIO_Abandon(bio);
    GWEN_BufferedIO_free(bio);
    GWEN_DB_Group_free(db);
    return;
  }

  bsize=GWEN_Buffer_GetUsedBytes(buf);
  err=GWEN_BufferedIO_WriteRawForced(bio,
                                     GWEN_Buffer_GetStart(buf),
                                     &bsize);
  if (!GWEN_Error_IsOk(err)) {
    DBG_ERROR_ERR(AQHBCI_LOGDOMAIN, err);
    GWEN_BufferedIO_Abandon(bio);
    GWEN_BufferedIO_free(bio);
    GWEN_DB_Group_free(db);
    return;
  }

  err=GWEN_BufferedIO_WriteLine(bio, "");
  if (!GWEN_Error_IsOk(err)) {
    DBG_ERROR_ERR(AQHBCI_LOGDOMAIN, err);
    GWEN_BufferedIO_Abandon(bio);
    GWEN_BufferedIO_free(bio);
    GWEN_DB_Group_free(db);
    return;
  }

  err=GWEN_BufferedIO_Close(bio);
  if (!GWEN_Error_IsOk(err)) {
    DBG_ERROR_ERR(AQHBCI_LOGDOMAIN, err);
    GWEN_BufferedIO_Abandon(bio);
    GWEN_BufferedIO_free(bio);
    GWEN_DB_Group_free(db);
    return;
  }

  GWEN_BufferedIO_free(bio);
  GWEN_DB_Group_free(db);
  DBG_INFO(AQHBCI_LOGDOMAIN, "Message logged");
}



const char *AH_Msg_GetExpectedSigner(const AH_MSG *msg){
  assert(msg);
  return msg->expectedSigner;
}



void AH_Msg_SetExpectedSigner(AH_MSG *msg, const char *s){
  assert(msg);
  free(msg->expectedSigner);
  if (s) msg->expectedSigner=strdup(s);
  else msg->expectedSigner=0;
}



const char *AH_Msg_GetExpectedCrypter(const AH_MSG *msg){
  assert(msg);
  return msg->expectedCrypter;
}



void AH_Msg_SetExpectedCrypter(AH_MSG *msg, const char *s){
  assert(msg);
  free(msg->expectedCrypter);
  if (s) msg->expectedCrypter=strdup(s);
  else msg->expectedCrypter=0;
}



void AH_Msg_SetPin(AH_MSG *msg, const char *s){
  assert(msg);
  if (msg->usedPin) {
    memset(msg->usedPin, 0, strlen(msg->usedPin));
    free(msg->usedPin);
  }
  if (s) msg->usedPin=strdup(s);
  else msg->usedPin=0;
}



const char *AH_Msg_GetPin(const AH_MSG *msg){
  assert(msg);
  return msg->usedPin;
}






